import asyncio
import sys
import os
import importlib.util
import platform
import threading
from functools import partial
import warnings
import contextlib
import socket
import logging
import textwrap

from importlib.metadata import distributions
import numpy as np
import pytest
import h5py
import aiohttp
from dask import distributed as dd
import dask
from scipy.sparse import csr_matrix
from sparseconverter import NUMPY, SCIPY_CSR, for_backend

from distributed.scheduler import Scheduler
import tornado.httpserver

import libertem.api as lt
from libertem.executor.inline import InlineJobExecutor
from libertem.executor.delayed import DelayedJobExecutor
from libertem.io.dataset import get_dataset_cls
from libertem.io.dataset.hdf5 import H5DataSet
from libertem.io.dataset.raw import RawFileDataSet
from libertem.io.dataset.memory import MemoryDataSet
from libertem.io.dataset.base import BufferedBackend, MMapBackend, DirectBackend
from libertem.executor.dask import DaskJobExecutor, cluster_spec
from libertem.executor.concurrent import ConcurrentJobExecutor
from libertem.common.threading import set_num_threads_env
from libertem.viz.base import Dummy2DPlot

from libertem.utils.devices import detect

from libertem.web.server import make_app, EventRegistry
from libertem.web.state import SharedState, ExecutorState
from libertem.web.event_bus import EventBus
from libertem.executor.base import AsyncAdapter
from libertem.common.async_utils import sync_to_async
from libertem.common.async_utils import adjust_event_loop_policy
from libertem.common.tracing import maybe_setup_tracing

maybe_setup_tracing("pytest")


# A bit of gymnastics to import the test utilities since this
# conftest.py file is shared between the doctests and unit tests
# and this file is outside the package
basedir = os.path.dirname(__file__)
location = os.path.join(basedir, "tests/utils.py")
spec = importlib.util.spec_from_file_location("utils", location)
utils = importlib.util.module_from_spec(spec)
spec.loader.exec_module(utils)

# Mitigation for https://github.com/dask/distributed/issues/6776
dask.config.set({"distributed.worker.profile.enabled": False})


def get_or_create_hdf5(tmpdir_factory, filename, *args, **kwargs):
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, filename)
    try:
        with h5py.File(filename, 'r') as f:
            yield f
    except OSError:
        with h5py.File(filename, "w") as f:
            f.create_dataset("data", *args, **kwargs)
        with h5py.File(filename, 'r') as f:
            yield f


@pytest.fixture(scope='session')
def hdf5(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test.h5", data=np.ones((5, 5, 16, 16)))


@pytest.fixture(scope='session')
def hdf5_2d(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-2d.h5", data=np.ones((16, 16)))


@pytest.fixture(scope='session')
def hdf5_3d(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-3d.h5", data=np.ones((17, 16, 16)))


@pytest.fixture(scope='session')
def hdf5_5d(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-5d.h5",
                                  data=np.ones((3, 5, 9, 16, 16)))


@pytest.fixture(scope='session')
def random_hdf5_large_sig(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-random.h5",
                                  data=np.random.randn(16, 16, 512, 512))


@pytest.fixture(scope='session')
def random_hdf5(tmpdir_factory):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-random.h5",
                                  data=np.random.randn(5, 5, 16, 16))


@pytest.fixture(scope='session')
def empty_hdf5(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/hdf5-empty.h5'
    try:
        with h5py.File(filename, 'r') as f:
            yield f
    except OSError:
        with h5py.File(filename, "w") as f:
            pass
        with h5py.File(filename, 'r') as f:
            yield f


@pytest.fixture(scope='session')
def hdf5_4d_data():
    data = np.random.randn(2, 10, 26, 26).astype("float32")
    yield data


@pytest.fixture(scope='session')
def hdf5_same_data_3d(tmpdir_factory, hdf5_4d_data):
    data = hdf5_4d_data.reshape((20, 26, 26))
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-reshape-3d.h5", data=data)


@pytest.fixture(scope='session')
def hdf5_same_data_4d(tmpdir_factory, hdf5_4d_data):
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-reshape-4d.h5", data=hdf5_4d_data)


@pytest.fixture(scope='session')
def hdf5_same_data_5d(tmpdir_factory, hdf5_4d_data):
    data = hdf5_4d_data.reshape((2, 2, 5, 26, 26))
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-reshape-5d.h5", data=data)


@pytest.fixture(scope='session')
def hdf5_same_data_1d_sig(tmpdir_factory, hdf5_4d_data):
    data = hdf5_4d_data.reshape((2, 10, 676))
    yield from get_or_create_hdf5(tmpdir_factory, "hdf5-test-reshape-1d-sig.h5", data=data)


@pytest.fixture(scope='session')
def raw_same_dataset_4d(tmpdir_factory, hdf5_4d_data):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-same-data-4d'
    hdf5_4d_data.tofile(str(filename))
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=(2, 10),
        dtype="float32",
        sig_shape=(26, 26),
    )
    ds.set_num_cores(4)
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture
def hdf5_ds_1(hdf5):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def hdf5_ds_2(random_hdf5):
    ds = H5DataSet(
        path=random_hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def hdf5_ds_large_sig(random_hdf5):
    ds = H5DataSet(
        path=random_hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def hdf5_ds_3d(hdf5_3d):
    ds = H5DataSet(
        path=hdf5_3d.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def hdf5_ds_2d(hdf5_2d):
    ds = H5DataSet(
        path=hdf5_2d.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def hdf5_ds_5d(hdf5_5d):
    ds = H5DataSet(
        path=hdf5_5d.filename, ds_path="data",
    )
    ds = ds.initialize(InlineJobExecutor())
    return ds


@pytest.fixture
def ds_complex():
    data = np.random.choice(
        a=[0, 1, 0+1j, 0-1j, 1+1j, 1-1j], size=(16, 16, 16, 16)
    ).astype('complex64')
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2,
    )
    return dataset


@pytest.fixture(scope='session')
def default_raw_data():
    return utils._mk_random(size=(16, 16, 128, 128), dtype='float32')


@pytest.fixture(scope='session')
def default_raw(tmpdir_factory, default_raw_data):
    lt_ctx = lt.Context(executor=InlineJobExecutor())
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-default'
    default_raw_data.tofile(str(filename))
    del default_raw_data
    ds = lt_ctx.load(
        "raw",
        path=str(filename),
        dtype="float32",
        nav_shape=(16, 16),
        sig_shape=(128, 128),
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(2)
    try:
        yield ds
    finally:
        lt_ctx.close()


@pytest.fixture(scope='session')
def prime_raw_data():
    return utils._mk_random(size=(13, 17, 19, 23), dtype='float32')


@pytest.fixture(scope='session')
def prime_raw(tmpdir_factory, prime_raw_data):
    lt_ctx = lt.Context(executor=InlineJobExecutor())
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-prime'
    prime_raw_data.tofile(str(filename))
    del prime_raw_data
    ds = lt_ctx.load(
        "raw",
        path=str(filename),
        dtype="float32",
        nav_shape=(13, 17),
        sig_shape=(19, 23),
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(2)
    try:
        yield ds
    finally:
        lt_ctx.close()


@pytest.fixture(scope='session')
def default_raw_asymm(tmpdir_factory, default_raw_data):
    lt_ctx = lt.Context(executor=InlineJobExecutor())
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-default'
    default_raw_data.tofile(str(filename))
    del default_raw_data
    ds = lt_ctx.load(
        "raw",
        path=str(filename),
        dtype="float32",
        nav_shape=(14, 17),
        sig_shape=(128, 128),
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(2)
    yield ds


@pytest.fixture(scope='session')
def buffered_raw(tmpdir_factory, default_raw_data):
    lt_ctx = lt.Context(executor=InlineJobExecutor())
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-buffered'
    default_raw_data.tofile(str(filename))
    del default_raw_data

    ds = lt_ctx.load(
        "raw",
        path=str(filename),
        dtype="float32",
        nav_shape=(16, 16),
        sig_shape=(128, 128),
        io_backend=BufferedBackend(),
    )
    yield ds


@pytest.fixture(scope='session')
def direct_raw(tmpdir_factory, default_raw_data):
    lt_ctx = lt.Context(executor=InlineJobExecutor())
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-direct'
    default_raw_data.tofile(str(filename))
    del default_raw_data

    ds = lt_ctx.load(
        "raw",
        path=str(filename),
        dtype="float32",
        nav_shape=(16, 16),
        sig_shape=(128, 128),
        io_backend=DirectBackend(),
    )
    yield ds


@pytest.fixture(scope='session')
def big_endian_raw(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-default-big-endian'
    data = utils._mk_random(size=(16, 16, 128, 128), dtype='>u2')
    data.tofile(str(filename))
    del data
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=(16, 16),
        dtype=">u2",
        sig_shape=(128, 128),
    )
    ds.set_num_cores(2)
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def large_raw_file(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-large-sparse'
    shape = (100, 100, 1216, 1216)
    dtype = np.uint16
    size = np.prod(np.int64(shape)) * np.dtype(dtype).itemsize
    if platform.system() == "Windows":
        os.system(f'FSUtil File CreateNew "{filename}" 0x{size:X}')
        os.system('FSUtil Sparse SetFlag "%s"' % filename)
        os.system(f'FSUtil Sparse SetRange "{filename}" 0 0x{size:X}')
    else:
        with open(filename, 'wb') as f:
            f.truncate(size)
        stat = os.stat(filename)
        if stat.st_blocks != 0:
            warnings.warn(
                f"Created file {filename} is not reported as "
                f"sparse: {stat}, blocks {stat.st_blocks}"
            )
    yield filename, shape, dtype


@pytest.fixture(scope='session')
def large_raw(large_raw_file):
    filename, shape, dtype = large_raw_file
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=shape[:2],
        dtype=dtype,
        sig_shape=shape[2:],
    )
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def medium_raw_file(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-medium-sparse'
    shape = (128, 128, 256, 256)
    dtype = np.uint16
    size = np.prod(np.int64(shape)) * np.dtype(dtype).itemsize
    if platform.system() == "Windows":
        os.system(f'FSUtil File CreateNew "{filename}" 0x{size:X}')
        os.system('FSUtil Sparse SetFlag "%s"' % filename)
        os.system(f'FSUtil Sparse SetRange "{filename}" 0 0x{size:X}')
    else:
        with open(filename, 'wb') as f:
            f.truncate(size)
        stat = os.stat(filename)
        if stat.st_blocks != 0:
            warnings.warn(
                f"Created file {filename} is not reported as "
                f"sparse: {stat}, blocks {stat.st_blocks}"
            )
    yield filename, shape, dtype


@pytest.fixture(scope='session')
def medium_raw(medium_raw_file):
    filename, shape, dtype = medium_raw_file
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=shape[:2],
        dtype=dtype,
        sig_shape=shape[2:],
        io_backend=MMapBackend()
    )
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def medium_raw_file_float32(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-medium-sparse'
    shape = (128, 128, 256, 256)
    dtype = np.float32
    size = np.prod(np.int64(shape)) * np.dtype(dtype).itemsize
    if platform.system() == "Windows":
        os.system(f'FSUtil File CreateNew "{filename}" 0x{size:X}')
        os.system('FSUtil Sparse SetFlag "%s"' % filename)
        os.system(f'FSUtil Sparse SetRange "{filename}" 0 0x{size:X}')
    else:
        with open(filename, 'wb') as f:
            f.truncate(size)
        stat = os.stat(filename)
        if stat.st_blocks != 0:
            warnings.warn(
                f"Created file {filename} is not reported as "
                f"sparse: {stat}, blocks {stat.st_blocks}"
            )
    yield filename, shape, dtype


@pytest.fixture(scope='session')
def medium_raw_float32(medium_raw_file_float32):
    filename, shape, dtype = medium_raw_file_float32
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=shape[:2],
        dtype=dtype,
        sig_shape=shape[2:],
        io_backend=MMapBackend()
    )
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def uint16_raw(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-test-default-uint16'
    data = utils._mk_random(size=(16, 16, 128, 128), dtype='uint16')
    data.tofile(str(filename))
    del data
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=(16, 16),
        dtype="uint16",
        sig_shape=(128, 128),
    )
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def raw_with_zeros(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/raw-with-zeros'
    data = np.zeros((16, 16, 128, 128), dtype='float32')
    data.tofile(str(filename))
    del data
    ds = RawFileDataSet(
        path=str(filename),
        nav_shape=(16, 16),
        dtype="float32",
        sig_shape=(128, 128),
    )
    ds.set_num_cores(2)
    ds = ds.initialize(InlineJobExecutor())
    yield ds


@pytest.fixture(scope='session')
def raw_data_8x8x8x8_path(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/8x8x8x8'
    data = utils._mk_random(size=(8, 8, 8, 8), dtype='float32')
    data.tofile(str(filename))
    del data
    yield str(filename)


@pytest.fixture(scope="session")
def mock_sparse_data():
    data = utils._mk_random((13, 17, 24, 19), array_backend=NUMPY)
    data_flat = data.reshape((13*17, 24*19))
    orig = for_backend(data_flat, SCIPY_CSR)
    return orig, data_flat


@pytest.fixture(scope="session")
def raw_csr_generated(mock_sparse_data: tuple[csr_matrix, np.ndarray], tmpdir_factory):
    orig, data_flat = mock_sparse_data
    datadir = tmpdir_factory.mktemp('raw_csr')
    name_indptr = str(datadir / 'indptr.raw')
    name_coords = str(datadir / 'coords.raw')
    name_values = str(datadir / 'values.raw')
    name_sidecar = str(datadir / 'sparse.toml')
    with open(name_sidecar, "w") as f:
        f.write(f"""
[params]
filetype = "raw_csr"
nav_shape = [13, 17]
sig_shape = [24, 19]

[raw_csr]
indptr_file = "indptr.raw"
indptr_dtype = "{str(orig.indptr.dtype)}"

indices_file = "coords.raw"
indices_dtype = "{str(orig.indices.dtype)}"

data_file = "values.raw"
data_dtype = "{str(orig.data.dtype)}"
""")
    orig.indptr.tofile(name_indptr)
    orig.indices.tofile(name_coords)
    orig.data.tofile(name_values)
    lt_ctx = lt.Context.make_with('inline')
    yield lt_ctx.load("raw_csr", path=name_sidecar)


@pytest.fixture(scope="session")
def raw_csr_generated_bigendian(
        mock_sparse_data: tuple[csr_matrix, np.ndarray], tmpdir_factory):
    orig, data_flat = mock_sparse_data
    dtype = np.dtype('>i4')
    datadir = tmpdir_factory.mktemp('raw_csr')
    name_indptr = str(datadir / 'indptr_bigendian.raw')
    name_coords = str(datadir / 'coords_bigendian.raw')
    name_values = str(datadir / 'values_bigendian.raw')
    name_sidecar = str(datadir / 'sparse_bigendian.toml')
    with open(name_sidecar, "w") as f:
        f.write(f"""
[params]
filetype = "raw_csr"
nav_shape = [13, 17]
sig_shape = [24, 19]

[raw_csr]
indptr_file = "indptr_bigendian.raw"
indptr_dtype = "{str(dtype)}"

indices_file = "coords_bigendian.raw"
indices_dtype = "{str(dtype)}"

data_file = "values_bigendian.raw"
data_dtype = "{str(dtype)}"
""")
    orig.indptr.astype(dtype).tofile(name_indptr)
    orig.indices.astype(dtype).tofile(name_coords)
    orig.data.astype(dtype).tofile(name_values)
    lt_ctx = lt.Context.make_with('inline')
    yield lt_ctx.load("raw_csr", path=name_sidecar)


@pytest.fixture(scope="session")
def raw_csr_generated_uint64(mock_sparse_data: tuple[csr_matrix, np.ndarray], tmpdir_factory):
    orig, data_flat = mock_sparse_data
    datadir = tmpdir_factory.mktemp('raw_csr')
    name_indptr = str(datadir / 'indptr_uint64.raw')
    name_coords = str(datadir / 'coords_uint64.raw')
    name_values = str(datadir / 'values_uint64.raw')
    name_sidecar = str(datadir / 'sparse_uint64.toml')
    with open(name_sidecar, "w") as f:
        f.write("""
[params]
filetype = "raw_csr"
nav_shape = [13, 17]
sig_shape = [24, 19]

[raw_csr]
indptr_file = "indptr_uint64.raw"
indptr_dtype = "<u8"

indices_file = "coords_uint64.raw"
indices_dtype = "<u8"

data_file = "values_uint64.raw"
data_dtype = "<u8"
""")
    orig.indptr.astype("<u8").tofile(name_indptr)
    orig.indices.astype("<u8").tofile(name_coords)
    orig.data.astype("<u8").tofile(name_values)
    lt_ctx = lt.Context.make_with('inline')
    yield lt_ctx.load("raw_csr", path=name_sidecar)


@pytest.fixture(scope='session')
def npy_datadir(tmpdir_factory):
    yield tmpdir_factory.mktemp('data_npy')


@pytest.fixture(scope='session')
def default_npy_filepath(npy_datadir):
    yield str(npy_datadir + '/test_default.npy')


@pytest.fixture()
def npy_random_array(npy_datadir):
    random_filename = str(npy_datadir + f'/array{np.random.randint(0, 1000)}.npy')
    ndim = np.random.randint(1, 6)
    shape = tuple(np.random.randint(1, 10) for _ in range(ndim))
    dtype = np.random.choice([np.float32, np.uint8, np.int64, np.complex128])
    array = np.empty(shape, dtype=dtype)
    np.save(random_filename, array)
    return random_filename, array


@pytest.fixture()
def npy_fortran_array(npy_datadir):
    random_filename = str(npy_datadir + f'/array{np.random.randint(0, 1000)}.npy')
    array = np.ones((55, 55), order='F')
    np.save(random_filename, array)
    return random_filename, array


@pytest.fixture(scope='session')
def npy_8x8x8x8_path(npy_datadir):
    filename = npy_datadir + '/8x8x8x8.npy'
    data = utils._mk_random(size=(8, 8, 8, 8), dtype='float32')
    np.save(str(filename), data)
    del data
    yield str(filename)


@pytest.fixture(scope='session')
def npy_8x8x8x8_ds(npy_8x8x8x8_path):
    lt_ctx = lt.Context.make_with('inline')
    ds = lt_ctx.load(
            'npy',
            path=npy_8x8x8x8_path,
        )
    ds.set_num_cores(2)
    yield ds


@pytest.fixture(scope='session')
def default_npy(default_npy_filepath, default_raw_data):
    lt_ctx = lt.Context.make_with('inline')
    filename = default_npy_filepath
    np.save(filename, default_raw_data)
    ds = lt_ctx.load(
        "npy",
        path=filename,
        sig_dims=2,
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(2)
    del default_raw_data
    yield ds


@pytest.fixture
def naughty_filename():
    '''
    Return a string with many special charaters that tests the limits
    of what the file system allows on that platform. This allows to stress-test
    globs or regular expressions applied to paths.
    '''
    system = platform.system()
    # See https://en.wikipedia.org/wiki/Filename#Comparison_of_filename_limitations
    if system == 'Windows':  # NTFS
        return "!§$&[%)(]=`´';,.#~ 🤪"
    elif system == 'Darwin':  # HFS+, APFS?
        return "!\"\\§$&[%)(]=?`´';,.# ~ * | < ** > 🤪"
    else:  # Linux, other Unix
        return "!\"\\§$&[%)(]=?`´':;,.# ~ * | < ** > 🤪"


@pytest.fixture
def scheduler_addr():
    return os.environ['DASK_SCHEDULER_ADDRESS']


@pytest.fixture
def dist_ctx(scheduler_addr):
    """
    This Context needs to have an external dask cluster running, with the following
    assumptions:

     - two workers: hostnames worker-1 and worker-2
     - one scheduler node
     - data availability TBD
    """
    executor = DaskJobExecutor.connect(scheduler_addr)
    with lt.Context(executor=executor) as ctx:
        yield ctx


@pytest.fixture
def ipy_ctx():
    import ipyparallel
    client = ipyparallel.Client()
    # wait for two engines: see also docker-compose.yml where the engines are started
    client.wait_for_engines(2)
    dask_client = client.become_dask()
    executor = DaskJobExecutor(client=dask_client, is_local=False)
    with lt.Context(executor=executor) as ctx:
        yield ctx


# Starting fresh distributed executors takes a lot of time and therefore
# they should be used repeatedly if possible.
# However, some benchmarks require a fresh distributed executor
# and running several Dask executors in parallel has led to lockups when closing
# in some instances.
# That means any shared executor should be shut down before a fresh one is started.
# For that reason we use a fixture with scope "class" and group
# tests in a class that should all use the same executor.
# That way we make sure the shared executor is torn down before any other test
# starts a new one.

# Different from the local_cluster_ctx fixture that only uses two CPUs and at most
# one GPU, this fixture starts a cluster for benchmarking under production condition that
# uses all available CPUs and GPUs. Furthermore, the LiberTEM Context and not only the
# Dask cluster is shared between functions.


@pytest.fixture(scope="class")
def shared_dist_ctx():
    print("start shared Context()")
    ctx = lt.Context()
    yield ctx
    print("stop shared Context()")
    ctx.close()


@pytest.fixture(scope="class")
def shared_dist_ctx_globaldask():
    # Sets default dask.distributed client
    # for integration testing
    print("start shared Context()")
    ctx = lt.Context.make_with('dask-make-default')
    yield ctx
    print("stop shared Context()")
    # Make sure everything is shut down
    ctx.executor.is_local = True
    ctx.close()


@pytest.fixture(autouse=True)
def auto_ctx(doctest_namespace):
    ctx = lt.Context(executor=InlineJobExecutor())
    doctest_namespace["ctx"] = ctx


@pytest.fixture(autouse=True)
def auto_ds(doctest_namespace):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32])
    doctest_namespace["dataset"] = dataset


@pytest.fixture(autouse=True)
def auto_libs(doctest_namespace):
    doctest_namespace["np"] = np


@pytest.fixture(autouse=True)
def auto_libertem(doctest_namespace):
    import libertem
    import libertem.utils
    import libertem.utils.generate
    import libertem.masks
    import libertem.api
    doctest_namespace["libertem"] = libertem
    doctest_namespace["libertem.utils"] = libertem.utils
    doctest_namespace["libertem.utils.generate"] = libertem.utils.generate
    doctest_namespace["libertem.masks"] = libertem.masks
    doctest_namespace["libertem.api"] = libertem.api


@pytest.fixture(autouse=True)
def auto_files(doctest_namespace, hdf5, default_raw):
    doctest_namespace["path_to_hdf5"] = hdf5.filename
    doctest_namespace["path_to_raw"] = default_raw._path


@pytest.fixture
def inline_executor():
    return InlineJobExecutor(debug=True, inline_threads=2)


@pytest.fixture
def delayed_executor():
    return DelayedJobExecutor()


@pytest.fixture
def lt_ctx(inline_executor):
    return lt.Context(executor=inline_executor, plot_class=Dummy2DPlot)


@pytest.fixture
def inline_executor_fast():
    return InlineJobExecutor(debug=False, inline_threads=2)


@pytest.fixture
def lt_ctx_fast(inline_executor_fast):
    return lt.Context(executor=inline_executor_fast, plot_class=Dummy2DPlot)


@pytest.fixture
async def async_executor(local_cluster_url):
    pool = AsyncAdapter.make_pool()
    sync_executor = await sync_to_async(
        partial(
            DaskJobExecutor.connect,
            scheduler_uri=local_cluster_url
        ),
        pool=pool,
    )
    executor = AsyncAdapter(wrapped=sync_executor, pool=pool)
    yield executor
    await executor.close()


@pytest.fixture
def dask_executor(local_cluster_url):
    executor = DaskJobExecutor.connect(local_cluster_url)
    yield executor
    executor.close()


@pytest.fixture
def concurrent_executor():
    executor = ConcurrentJobExecutor.make_local()
    yield executor
    executor.close()


@pytest.fixture
def delayed_ctx(delayed_executor):
    return lt.Context(executor=delayed_executor, plot_class=Dummy2DPlot)


@pytest.fixture
async def http_client():
    # FIXME: maybe re-scope to module, but would also need
    # adjusted event_loop scope. if we have many API tests
    # maybe reconsider.
    # The timeout needs to be this high to acommodate overloaded
    # CI environments, or otherwise oversubscribed systems
    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(total=120)
    ) as session:
        yield session


@pytest.fixture
async def shared_state():
    event_bus = EventBus()
    executor_state = ExecutorState(event_bus=event_bus)
    yield SharedState(executor_state=executor_state)
    executor_state.shutdown()


class ServerThread(threading.Thread):
    def __init__(self, port, shared_state, token, *args, **kwargs):
        super().__init__(name='LiberTEM-background', *args, **kwargs)
        self.stop_event = threading.Event()
        self.start_event = threading.Event()
        self.port = port
        self.shared_state = shared_state
        self.token = token
        self.loop = None

    async def stop(self):
        self.server.stop()
        await self.server.close_all_connections()
        exstate = self.shared_state.executor_state
        if exstate.have_executor() and not exstate.is_snoozing():
            executor = await exstate.get_executor()
            await executor.close()
        self.loop.stop()

    async def wait_for_stop(self):
        """
        background task to periodically check if the main thread wants
        us to stop
        """
        try:
            while True:
                if self.stop_event.is_set():
                    await self.stop()
                    break
                await asyncio.sleep(0.1)
        except Exception:
            await self.stop()
            raise

    def run(self):
        try:
            adjust_event_loop_policy()
            self.loop = loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.set_debug(True)

            event_registry = EventRegistry()
            app = make_app(event_registry, self.shared_state, self.token)
            self.server = tornado.httpserver.HTTPServer(app)
            self.server.listen(address="127.0.0.1", port=self.port)
            # self.shared_state.set_server(self.server)

            asyncio.ensure_future(self.wait_for_stop())
            self.start_event.set()
            loop.run_forever()
        finally:
            self.loop.stop()


@contextlib.contextmanager
def common_server_startup(unused_tcp_port_factory, shared_state, token):
    """
    start a LiberTEM API server on a unused port
    """
    port = unused_tcp_port_factory()

    print(f"starting server at port {port}")
    thread = ServerThread(port, shared_state, token, daemon=True)
    thread.start()
    assert thread.start_event.wait(timeout=5), "server thread failed to start"
    yield port
    print(f"stopping server at port {port}")
    thread.stop_event.set()
    thread.join(timeout=15)
    if thread.is_alive():
        raise RuntimeError("thread did not stop in the given timeout")


@pytest.fixture(scope="function")
def default_token():
    token = "something_random"
    return token


@pytest.fixture(scope="function")
def server_port(unused_tcp_port_factory, shared_state, default_token):
    with common_server_startup(unused_tcp_port_factory, shared_state, token=default_token) as port:
        yield port


@pytest.fixture(scope="function")
async def base_url(server_port):
    return "http://127.0.0.1:%d" % server_port


@pytest.fixture(scope="function")
async def base_url_no_token(unused_tcp_port_factory, shared_state):
    with common_server_startup(
            unused_tcp_port_factory, shared_state, token=None
    ) as server_port:
        yield "http://127.0.0.1:%d" % server_port


def find_unused_port():
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('127.0.0.1', 0))
        return sock.getsockname()[1]


if sys.platform.startswith("win"):
    # Py3.8+ on Windows need some help with the event loop policy:
    @pytest.fixture(scope="session", autouse=True)
    def event_loop_policy():
        # see also: libertem.common.async_utils.adjust_event_loop_policy
        try:
            from asyncio import (
                WindowsProactorEventLoopPolicy,
                WindowsSelectorEventLoopPolicy,
            )
        except ImportError:
            return asyncio.get_event_loop_policy()
        else:
            # FIXME this might fail if the event loop policy has been overridden by something
            # incompatible that is not a WindowsProactorEventLoopPolicy. Dask.distributed creates
            # a custom event loop policy, for example. Fortunately that one should be compatible.
            # Better would be "duck typing", i.e. to check if the required add_reader() method
            # of the loop throws a NotImplementedError.
            if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                # fallback to the pre-3.8 default of Selector
                return WindowsSelectorEventLoopPolicy()


@pytest.fixture(scope='session')
def local_cluster_url():
    """
    Shared dask cluster, can be used repeatedly by different executors.

    This allows numba caching across tests, without sharing the executor,
    for example
    """
    cluster_port = find_unused_port()
    devices = detect()
    spec = cluster_spec(
        # Only use at most 2 CPUs and 1 GPU
        cpus=devices['cpus'][:2],
        cudas=devices['cudas'][:1],
        has_cupy=devices['has_cupy']
    )

    cluster_kwargs = {
        'silence_logs': logging.WARN,
        'scheduler': {
            'cls': Scheduler,
            'options': {'port': cluster_port},
        },
    }

    # use the context manager of SpecCluster here to properly
    # shut down the LoopRunner in case of older distributed versions:
    with dd.SpecCluster(
        workers=spec,
        **(cluster_kwargs or {})
    ) as cluster:
        with set_num_threads_env(1, set_numba=False):
            client = dd.Client(cluster, set_as_default=False)
            client.wait_for_workers(len(spec))
            client.close()

        yield 'tcp://localhost:%d' % cluster_port


@pytest.fixture(scope='function')
def local_cluster_ctx(dask_executor):
    return lt.Context(executor=dask_executor)


@pytest.fixture
def concurrent_ctx(concurrent_executor):
    return lt.Context(executor=concurrent_executor)


@pytest.fixture(scope='session')
def pipelined_ctx():
    # limit the number of CPU workers started by default
    with lt.Context.make_with('pipelined', cpus=2) as ctx:
        yield ctx


@pytest.fixture
def points():
    return np.array([
        (0, 0),
        (0, 1),
        (1, 0),
        (1, 1),
        (0, -1),
        (-1, 0),
        (-1, -1)
    ])


@pytest.fixture(scope="function")
def standard_bad_ds_params():
    return [
        dict(nav_shape=(-1, 2)),
        dict(nav_shape=4),
        dict(sig_shape=(3, 17)),
        dict(sig_shape=(103, -83)),
        dict(sync_offset=180000000),
        dict(io_backend="doesn't exist"),
    ]


@pytest.fixture()
def ds_params_tester(lt_ctx, bad_params_logger):
    def _test_bad_params(ds_key, *args, **params):
        ds_class = get_dataset_cls(ds_key)
        kwarg_repr = (
            textwrap.shorten(f'{k}={repr(v)}', 30, placeholder="...")
            for k, v in params.items()
        )
        stub = (
            f"{ds_class.__name__}(..., {', '.join(kwarg_repr)})"
        )
        try:
            lt_ctx.load(ds_key, *args, **params)
        except Exception as e:
            bad_params_logger.append(f'{stub}\n  RAISED {type(e).__name__}("{e}")')
            return False
        bad_params_logger.append(f'{stub}\n  DID NOT raise')
        return True
    return _test_bad_params


@pytest.fixture(scope="session")
def bad_params_logger():
    lines = []
    yield lines
    print("\n".join(lines))


@pytest.hookimpl(hookwrapper=True)
def pytest_benchmark_generate_json(config, benchmarks, include_data, machine_info, commit_info):
    machine_info["freeze"] = [
        (d.metadata['Name'], d.version)
        for d in distributions()
    ]
    yield


def pytest_collectstart(collector):
    # nbval: ignore some output types
    if collector.fspath and collector.fspath.ext == '.ipynb':
        collector.skip_compare += 'text/html', 'application/javascript', 'stderr',



from enum import IntEnum
from typing import NamedTuple, Optional, Union
from typing_extensions import Literal

import numpy as np
from sparseconverter import CUPY_BACKENDS

from libertem import masks
from libertem.corrections import coordinates
from libertem.udf.masks import ApplyMasksEngine
from libertem.common.container import MaskContainer
from libertem.common.math import prod
from libertem.udf.base import UDF


class RegressionOptions(IntEnum):
    """
    Possible values of the regression functionality
    implemented in :class:`~libertem.udf.com.CoMUDF`
    """
    NO_REGRESSION = -1
    SUBTRACT_MEAN = 0
    SUBTRACT_LINEAR = 1


# Needed because poor enum support in mypy?
RegressionOptionsT = Union[np.ndarray, Literal[-1, 0, 1]]


class CoMParams(NamedTuple):
    """
    Used to instantiate :class:`~libertem.udf.com.CoMUDF`.

    See :meth:`~libertem.udf.com.CoMUDF.with_params`
    for descriptions of each parameter. In most cases use this
    method rather than instantiating this class directly.
    """
    cy: Optional[float] = None
    cx: Optional[float] = None
    r: float = float('inf')
    ri: Union[float, None] = 0.
    scan_rotation: float = 0.
    flip_y: bool = False
    regression: RegressionOptionsT = RegressionOptions.NO_REGRESSION


def com_masks_factory(detector_y, detector_x, cy, cx, r):
    def disk_mask():
        return masks.circular(
            centerX=cx, centerY=cy,
            imageSizeX=detector_x,
            imageSizeY=detector_y,
            radius=r,
        )

    return [
        disk_mask,
        lambda: masks.gradient_y(
            imageSizeX=detector_x,
            imageSizeY=detector_y,
        ) * disk_mask(),
        lambda: masks.gradient_x(
            imageSizeX=detector_x,
            imageSizeY=detector_y,
        ) * disk_mask(),
    ]


def com_masks_generic(detector_y, detector_x, base_mask_factory):
    """
    Create a CoM mask stack with a generic selection mask factory

    Parameters
    ----------
    detector_y : int
        The detector height
    detector_x : int
        The detector width
    base_mask_factory : () -> np.array
        A factory function for creating the selection mask

    Returns
    -------
    List[Function]
        The mask stack as a list of factory functions
    """
    return [
        base_mask_factory,
        lambda: masks.gradient_y(
            imageSizeX=detector_x,
            imageSizeY=detector_y,
        ) * base_mask_factory(),
        lambda: masks.gradient_x(
            imageSizeX=detector_x,
            imageSizeY=detector_y,
        ) * base_mask_factory(),
    ]


def center_shifts(img_sum, img_y, img_x, ref_y, ref_x):
    x_centers = np.divide(img_x, img_sum, where=img_sum != 0)
    y_centers = np.divide(img_y, img_sum, where=img_sum != 0)
    x_centers[img_sum == 0] = ref_x
    y_centers[img_sum == 0] = ref_y
    x_centers -= ref_x
    y_centers -= ref_y
    return (y_centers, x_centers)


def apply_correction(
    y_centers, x_centers, scan_rotation, flip_y, forward=True,
) -> tuple[np.ndarray, np.ndarray]:
    shape = y_centers.shape
    if flip_y:
        transform = coordinates.flip_y()
    else:
        transform = coordinates.identity()
    # Transformations are applied right to left
    transform = coordinates.rotate_deg(scan_rotation) @ transform
    y_centers = y_centers.reshape(-1)
    x_centers = x_centers.reshape(-1)
    if not forward:
        transform = np.linalg.inv(transform)
    y_transformed, x_transformed = transform @ (y_centers, x_centers)
    y_transformed = y_transformed.reshape(shape)
    x_transformed = x_transformed.reshape(shape)
    return (y_transformed, x_transformed)


def divergence(y_centers, x_centers):
    return np.gradient(y_centers, axis=0) + np.gradient(x_centers, axis=1)


def curl_2d(y_centers, x_centers):
    # https://en.wikipedia.org/wiki/Curl_(mathematics)#Usage
    # DFy/dx - dFx/dy
    # axis 0 is y, axis 1 is x
    return np.gradient(y_centers, axis=1) - np.gradient(x_centers, axis=0)


def magnitude(y_centers, x_centers):
    return np.sqrt(y_centers**2 + x_centers**2)


def coordinate_check(y_centers, x_centers, roi=None):
    '''
    Calculate the RMS curl as a function of :code:`scan_rotation` and :code:`flip_y`.

    The curl for a purely electrostatic field is zero. That means
    the correct settings for :code:`scan_rotation` and :code:`flip_y` should
    minimize the RMS curl for atomic resolution STEM of non-magnetic specimens
    along a high-symmetry zone axis.

    Parameters
    ----------
    y_centers, x_centers : numpy.ndarray
        2D arrays with y and x component of the center of mass shift for each
        scan position, as returned by :meth:`center_shifts` or
        :meth:`apply_correction`
    roi : Optional[numpy.ndarray]
        Selector for values to consider in the statistics, compatible
        with indexing an array with the shape of y_centers and x_centers.
        By default, everything except the last row and last column are used
        since these contain artefacts.

    Returns
    -------
    (straight, flipped)
        Root mean square of the curl as a function of :code:`scan_rotation` from
        0 to 359 degrees in steps of one, with :code:`flip_y=False` (straight)
        and code:`flip_y=True` (flipped).
    '''
    straight = np.zeros(360)
    flipped = np.zeros(360)
    if roi is None:
        # The last row and column contain artifacts
        roi = (slice(0, -1), slice(0, -1))
    for angle in range(360):
        for flip_y in (True, False):
            y_transformed, x_transformed = apply_correction(
                y_centers, x_centers, scan_rotation=angle, flip_y=flip_y
            )
            curl = curl_2d(y_transformed, x_transformed)
            result = np.sqrt(np.mean(curl[roi]**2))
            if flip_y:
                flipped[angle] = result
            else:
                straight[angle] = result
    return (straight, flipped)


class GuessResult(NamedTuple):
    """
    Container for CoM parameters inferred from results,
    which can be passed to :meth:`~libertem.udf.com.CoMUDF.with_params`
    to create a new :class:`~libertem.udf.com.CoMUDF` better-adapted
    to the data.

    Return type of :func:`~libertem.udf.com.guess_corrections`
    """
    scan_rotation: int
    flip_y: bool
    cy: float
    cx: float


def guess_corrections(
    y_centers: np.ndarray,
    x_centers: np.ndarray,
    roi: Optional[Union[np.ndarray, tuple[slice, ...]]] = None,
) -> GuessResult:
    '''
    Guess values for center offset (:code:`(cy, cx)`), :code:`scan_rotation` and
    :code:`flip_y` from existing CoM data.

    This function can generate a CoM parameter guess for atomic resolution 4D STEM data
    by using the following assumptions:

    * The field is purely electrostatic, i.e. the RMS curl should be minimized
    * There is no net field over the field of view and no descan error,
      i.e. the mean deflection is zero.
    * Atomic resolution STEM means that the divergence will be negative at atom columns
      and consequently the histogram of divergence will have a stronger tail towards
      negative values than towards positive values.

    If any corrections were applied when generating the input data please note that
    the new parameters should be applied relative to these previous values. In particular,
    the values for :code:`(cy, cx)` returned by this function are the means of the inputs
    :code:`y_centers` and :code:`x_centers`. If these inputs are CoM-shifts relative
    to a given position, then the guessed correction will also be relative to this same
    position.

    Parameters
    ----------
    y_centers, x_centers : numpy.ndarray
        2D arrays with y and x component of the center of mass for each scan position
    roi : Optional[numpy.ndarray]
        Selector for values to consider in the statistics, compatible
        with indexing an array with the shape of y_centers and x_centers.
        By default, everything except the last row and last column are used
        since these contain artefacts.

    Examples
    --------
    >>> # Perform default, whole-frame CoM on a dataset
    >>> udf = CoMUDF()
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> # Pass the `raw_com` result to this function
    >>> y_centers = result['raw_com'].data[..., 0]
    >>> x_centers = result['raw_com'].data[..., 1]
    >>> guess_result = guess_corrections(y_centers, x_centers)
    >>> # Create a new CoMUDF with the guessed parameters
    >>> param_udf = CoMUDF.with_params(**guess_result._asdict())
    >>> result_corrected = ctx.run_udf(dataset=dataset, udf=param_udf)

    Returns
    -------
    GuessResult : computed corrections
    '''
    if roi is None:
        # The last row and column contain artefacts
        roi = (slice(0, -1), slice(0, -1))
    straight, flipped = coordinate_check(y_centers, x_centers, roi=roi)
    # The one with lower minima is the correct one
    flip_y = bool(np.min(flipped) < np.min(straight))
    if flip_y:
        angle = np.argmin(flipped)
    else:
        angle = np.argmin(straight)

    corrected_y, corrected_x = apply_correction(
        y_centers, x_centers, scan_rotation=angle, flip_y=flip_y
    )
    # There are two equivalent angles that minimize RMS curl since a 180°
    # rotation inverts the coordinates and just flips the sign for divergence
    # and curl. To distinguish the two, the distribution of the divergence is
    # analyzed. With negative electrons and positive nuclei, the beam is
    # deflected towards the nuclei and the divergence is negative there. Since
    # the beam is deflected most strongly near the nuclei, the histogram should
    # have more values at the negative end of the range than at the positive
    # end.
    div = divergence(corrected_y, corrected_x)[roi]
    all_range = np.maximum(-np.min(div), np.max(div))
    hist, bins = np.histogram(div, range=(-all_range, all_range), bins=5)
    polarity_off = np.sum(hist[:1]) < np.sum(hist[-1:])
    if polarity_off:
        angle += 180
    if angle > 180:
        angle -= 360
    return GuessResult(
        scan_rotation=int(angle),
        flip_y=flip_y,
        cy=np.mean(y_centers[roi]),
        cx=np.mean(x_centers[roi])
    )


class CoMUDF(UDF):
    """
    Perform centre-of-mass analysis on the dataset

    To parametrise the CoM calculation, use the constructor
    :meth:`CoMUDF.with_params`.

    This replicates the functionality of
    :func:`~libertem.api.Context.create_com_analysis`
    but in UDF form, making it compatible with live processing and more easily
    sub-classable.

    .. versionadded:: 0.12.0

    The result buffers of this UDF, all of type 'nav', are as follows:

    * raw_com, coordinate (y, x)
        The centre-of-mass in the pixel coordinate system of the
        frame, subject to the mask defined by the parameters, if any.

    * raw_shifts, vector (dy, dx)
        The raw shift in centre-of-mass in axis-aligned pixels, relative
        to the parameters cy and cx, either as-supplied or from
        the frame centre, by default.

    * field, vector (2,)
        The transformed shift in the centre-of-mass in pixels, subject
        to any corrections (scan rotation, detector flip)

    * field_y, scalar
        y-component of field, as given above.

    * field_x, scalar
        x-component of field, as given above.

    * magnitude, scalar
        The magnitude of the field buffer, equivalent to ||raw_shifts||_2

    * divergence, scalar
        The discrete divergence of the field buffer

    * curl, scalar
        The discrete curl of the field buffer

    * regression, (3, 2) matrix
        The regression parameters used to flatten the field:
        ((y, x), (dy/dy, dx/dy), (dy/dx, dx/dx))

    .. note::
        The implementation of the results 'divergence' and 'curl' differ slightly
        from :code:`COMAnalysis` in that here results for pixels outside of an ROI
        will always be NaN. The :code:`COMAnalysis` implementation could provide
        non-NaN values even in :code:`ROI == False` pixels due to the behaviour
        of :code:`np.gradient`.

    Parameters
    ----------
    com_params : CoMParams
        A :class:`CoMParams` instance containing the parameters
        for this UDF. By default a :code:`CoMParams` instance is created
        which performs whole-frame CoM giving shifts relative to the
        frame centre with no corrections.

    Examples
    --------
    >>> # CoM in a disk r=4. centred on (7.2, 6.8)
    >>> udf = CoMUDF.with_params(cy=7.2, cx=6.8, r=4.)
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> result["raw_shifts"].data.shape
    (16, 16, 2)

    >>> # Whole-frame CoM corrected to have zero-mean shift
    >>> udf = CoMUDF.with_params(regression=0)
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> result["raw_shifts"].data.shape
    (16, 16, 2)
    """
    def __init__(self, com_params: CoMParams = CoMParams()):
        super().__init__(com_params=com_params)

    @classmethod
    def with_params(
        cls,
        *,
        cy: Optional[float] = None,
        cx: Optional[float] = None,
        r: float = float('inf'),
        ri: float = 0.,
        scan_rotation: float = 0.,
        flip_y: bool = False,
        regression: RegressionOptionsT = RegressionOptions.NO_REGRESSION,
    ):
        """
        Returns an instantiated CoMUDF with a given set of parameters

        Parameters
        ----------
        cy : Optional[float], by default None
            Vertical-Centre of the mask applied to the frame, if any, and the
            reference point from which vertical CoM-shifts are calculated. If
            None, cy is set to the frame centre at runtime.
        cx : Optional[float], by default None
            Horizontal-Centre of the mask applied to the frame, if any, and the
            reference point from which horizontal CoM-shifts are calculated. If
            None, cx is set to the frame centre at runtime.
        r : float, by default :code:`float('inf')`
            (Outer) Radius of the disk mask around cy/cx to restrict the CoM
            calculation. The default value is :code:`float('inf')` which is
            equivalent to performing whole-frame CoM with the given origin
            cy/cx.
        ri : float, by default 0.
            (Inner) Radius of the ring mask around cy/cx to exclude from the CoM
            calculation. If left as 0., no inner disk is excluded and the CoM is
            calculated within a complete disk of radius r.
        scan_rotation : float, by default 0.
            Scan rotation in degrees. The optics of an electron microscope can
            rotate the image. Furthermore, scan generators may allow scanning in
            arbitrary directions. This means that the x and y coordinates of the
            detector image are usually not parallel to the x and y scan
            coordinates. For interpretation of center of mass shifts, however,
            the shift vector in detector coordinates has to be put in relation
            to the position on the sample. The :code:`scan_rotation` parameter
            can be used to rotate the detector coordinates to match the scan
            coordinate system. A positive value rotates the displacement vector
            clock-wise. That means if the detector seems rotated to the right
            relative to the scan, this value should be negative to counteract
            this rotation.
        flip_y : bool, by default False
            Flip the Y coordinate. Some detectors, namely Quantum Detectors
            Merlin, may have pixel (0, 0) at the lower left corner. This has to
            be corrected to get the sign of the y shift as well as curl and
            divergence right.
        regression : Union[np.ndarray, Literal[-1, 0, 1]], by default -1
            Regression to a background CoM field used to flatten the results

            Can be an ndarray with shape (3, 2) which specifies a correction
            to subtract in the form: :code:`[[y, x], [dy/dy, dx/dy], [dy/dx, dx/dx]]`.

            Otherwise a member of the :class:`~libertem.udf.com.RegressionOptions`
            enum or an integer in :code:`(-1, 0, 1)`. This specifies the regression
            order to calculate and apply to the results. With the following effects:

              - :code:`RegressionOptions.NO_REGRESSION` or :code:`-1`: No regression
              - :code:`RegressionOptions.SUBTRACT_MEAN` or :code:`0`: Subtract the mean
                to compensate for center offset.
              - :code:`RegressionOptions.SUBTRACT_LINEAR` or :code:`1`: Subtract the mean
                and a linear regression, to compensate for center offset and descan
                error.

            The regression that was used is available in the :code:`'regression'`
            result buffer.

        Returns
        -------
        CoMUDF : libertem.udf.com.CoMUDF
            the UDF with the provided parameters


        .. note::
            Using a regression correction calculated on the data itself may
            distort valid results, such as the effect of long-range fields.
            Furthermore, it may yield erronerous results due to aliasing on
            periodic structures or boundary effects. Subtracting a reference
            measurement or regression on a reference measurement instead of a
            regression on the data may give more reliable results.
        """
        if ri >= r:
            raise ValueError('Inner radius must be less than outer radius for annular CoM')
        return cls(
            com_params=CoMParams(
                cy=cy, cx=cx, r=r, ri=ri,
                scan_rotation=scan_rotation, flip_y=flip_y,
                regression=regression,
            )
        )

    def get_backends(self):
        return self.BACKEND_ALL

    def get_result_buffers(self):
        dtype = np.result_type(self.meta.input_dtype, np.float32)
        return {
            'raw_mask_result': self.buffer(
                kind='nav', dtype=dtype, extra_shape=(3, ), where='device', use='private'
            ),
            'raw_com': self.buffer(
                kind='nav', dtype=dtype, extra_shape=(2, ), use='result_only'
            ),
            'raw_shifts': self.buffer(
                kind='nav', dtype=dtype, extra_shape=(2, ), use='result_only'
            ),
            'field': self.buffer(
                kind='nav', dtype=dtype, extra_shape=(2, ), use='result_only'
            ),
            'field_y': self.buffer(
                kind='nav', dtype=dtype, use='result_only'
            ),
            'field_x': self.buffer(
                kind='nav', dtype=dtype, use='result_only'
            ),
            'magnitude': self.buffer(
                kind='nav', dtype=dtype, use='result_only'
            ),
            'divergence': self.buffer(
                kind='nav', dtype=dtype, use='result_only'
            ),
            'curl': self.buffer(
                kind='nav', dtype=dtype, use='result_only'
            ),
            'regression': self.buffer(
                kind='single', extra_shape=(3, 2), dtype=np.float64, use='result_only'
            ),
        }

    def get_params(self) -> CoMParams:
        sig_shape = tuple(self.meta.dataset_shape.sig)
        cy = self.params.com_params.cy
        if cy is None:
            cy = sig_shape[0] // 2

        cx = self.params.com_params.cx
        if cx is None:
            cx = sig_shape[1] // 2

        r = self.params.com_params.r
        ri = self.params.com_params.ri
        scan_rotation = self.params.com_params.scan_rotation
        flip_y = self.params.com_params.flip_y
        regression = self.params.com_params.regression

        cp = CoMParams(
            cy=cy, cx=cx, r=r, ri=ri, scan_rotation=scan_rotation, flip_y=flip_y,
            regression=regression,
        )
        return cp

    def get_task_data(self):
        sig_shape = tuple(self.meta.dataset_shape.sig)
        com_params = self.get_params()
        if len(sig_shape) != 2:
            raise ValueError('CoMUDF only works with 2D sig shape.')
        if len(self.meta.dataset_shape.nav) != 2:
            raise ValueError('CoMUDF only works with 2D nav shape.')

        if com_params.ri is None or np.isclose(com_params.ri, 0.):
            mask_factory = com_masks_factory(
                detector_y=sig_shape[0],
                detector_x=sig_shape[1],
                cx=com_params.cx,
                cy=com_params.cy,
                r=com_params.r,
            )
        else:
            mask_factory = com_masks_generic(
                detector_y=sig_shape[0],
                detector_x=sig_shape[1],
                base_mask_factory=lambda: masks.ring(
                    imageSizeY=sig_shape[0],
                    imageSizeX=sig_shape[1],
                    centerY=com_params.cy,
                    centerX=com_params.cx,
                    radius=com_params.r,
                    radius_inner=com_params.ri,
                )
            )
        if self.meta.array_backend in CUPY_BACKENDS:
            backend = self.BACKEND_CUPY
        else:
            backend = self.BACKEND_NUMPY

        masks_container = MaskContainer(
            mask_factories=mask_factory, dtype=np.float32, use_sparse=False,
            count=3, backend=backend
        )
        return {
            'com_params': com_params,
            'engine': ApplyMasksEngine(masks=masks_container, meta=self.meta, use_torch=True)
        }

    def process_tile(self, tile):
        engine: ApplyMasksEngine = self.task_data.engine
        raw_result = engine.process_tile(tile)
        self.results.raw_mask_result[:] += self.forbuf(
            raw_result, self.results.raw_mask_result
        )

    def get_field_results(self, field_y, field_x):
        '''
        To be overwritten in subclasses, such as for iCoM results
        in https://github.com/LiberTEM/LiberTEM-iCoM
        '''
        mag = magnitude(
            y_centers=field_y, x_centers=field_x
        )
        div = divergence(y_centers=field_y, x_centers=field_x)
        curl = curl_2d(y_centers=field_y, x_centers=field_x)
        return {
            'magnitude': mag,
            'divergence': div,
            'curl': curl,
        }

    def get_regression(self, field: np.ndarray, valid_mask: np.ndarray):
        inp = None
        result = np.zeros((3, 2))

        com_params = self.get_params()

        def get_inp():
            inp = np.ones(field.shape[:-1] + (3, ))
            y, x = np.ogrid[:field.shape[0], :field.shape[1]]
            inp[..., 1] = y
            inp[..., 2] = x
            return inp

        if isinstance(com_params.regression, int):
            if com_params.regression == -1:
                pass
            elif com_params.regression == 0:
                valid_field = field[valid_mask]
                result[0] = np.mean(valid_field, axis=0)
            elif com_params.regression == 1:
                inp = get_inp()
                valid_field = field[valid_mask]
                res = np.linalg.lstsq(inp[valid_mask], valid_field, rcond=None)
                result[:] = res[0]
            else:
                raise ValueError(f'Unrecognized regression option {com_params.regression}')
        else:
            regression = np.array(com_params.regression)
            if regression.shape != (3, 2):
                raise ValueError(
                    f"Regression parameter {com_params.regression} "
                    "doesn't have required shape (3, 2)."
                )
            result[:] = regression
        has_lin_regression = not np.allclose(result[1:], 0)
        if has_lin_regression and inp is None:
            inp = get_inp()
        if not has_lin_regression:
            inp = None
        return result, inp

    def apply_mean_regression(self, regression, field_inout, valid_mask):
        field_inout[valid_mask] -= regression[0]

    def apply_lin_regression(self, regression, inp, field_inout, valid_mask):
        valid_inp = inp[valid_mask]
        # fancy indexing flattens the nav shape, so it fits
        corr = valid_inp @ regression
        field_inout[valid_mask] -= corr

    def get_results(self):
        com_params = self.get_params()
        raw_mask_result = self.results.get_buffer('raw_mask_result')
        raw_shifts = center_shifts(
            img_sum=raw_mask_result.data[..., 0],
            img_y=raw_mask_result.data[..., 1],
            img_x=raw_mask_result.data[..., 2],
            ref_y=com_params.cy,
            ref_x=com_params.cx,
        )
        # CoM in the coordinate system of the frame
        # Ideally should calculate this first then shift
        # the centre by cy/cx, this is done just for backwards
        # compatibility of center_shifts
        raw_com = (
            raw_shifts[0].copy() + com_params.cy,
            raw_shifts[1].copy() + com_params.cx,
        )
        field = apply_correction(
            y_centers=raw_shifts[0],
            x_centers=raw_shifts[1],
            scan_rotation=com_params.scan_rotation,
            flip_y=com_params.flip_y,
        )
        roi = self.meta.roi

        raw_shifts = np.moveaxis(np.array(raw_shifts), 0, -1)
        raw_com = np.moveaxis(np.array(raw_com), 0, -1)
        field = np.moveaxis(np.array(field), 0, -1)
        nav_size = prod(self.meta.dataset_shape.nav)

        # Only apply the regression to the valid
        # part of the results:
        valid_mask = self.meta.get_valid_nav_mask(full_nav=True).reshape(
            tuple(self.meta.dataset_shape.nav)
        )
        regression, inp = self.get_regression(field, valid_mask=valid_mask)

        # inp not None indicates that we should apply a linear regression
        if inp is not None:
            self.apply_lin_regression(regression, inp, field, valid_mask)
        elif not np.allclose(regression[0], 0):
            self.apply_mean_regression(regression, field, valid_mask)

        results = {
            'raw_shifts': raw_shifts,
            'raw_com': raw_com,
            'field': field,
            'field_y': field[..., 0],
            'field_x': field[..., 1],
            'regression': regression.astype(np.float64),
        }

        results.update(self.get_field_results(field_y=field[..., 0], field_x=field[..., 1]))

        # Since we have circumvented the ROI mechanism by getting a raw buffer
        # so that we can calculate gradients, we have to filter and reshape the
        # results manually to bring them into the expected format.
        buffers = self.get_result_buffers()
        if self.meta.roi is not None:
            for key, buf in buffers.items():
                if buf.kind == 'nav' and key in results:
                    results[key] = results[key][roi]
        else:
            for key, buf in buffers.items():
                if buf.kind == 'nav' and key in results:
                    results[key] = results[key].reshape((nav_size, -1))
        return results



import os
import sys
import random
from pathlib import Path
import glob

import numpy as np
import pytest
from PIL import Image
import cloudpickle

from libertem.io.dataset.tvips import TVIPSDataSet, get_filenames
from libertem.udf.raw import PickUDF
from libertem.common import Shape
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.base import (
    TilingScheme, BufferedBackend, MMapBackend, DirectBackend
)
from libertem.common.buffers import reshaped_view

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

TVIPS_TESTDATA_PATH = os.path.join(get_testdata_path(), 'TVIPS', 'rec_20200623_080237_000.tvips')
HAVE_TVIPS_TESTDATA = os.path.exists(TVIPS_TESTDATA_PATH)

TVIPS_REFERENCE_DATA_PATTERN = os.path.join(
    get_testdata_path(), 'TVIPS', 'rec_20200623_080237_*.tif'
)

needsdata = pytest.mark.skipif(not HAVE_TVIPS_TESTDATA, reason="need .tvips testdata")  # NOQA


def get_reference_files():
    files = glob.glob(TVIPS_REFERENCE_DATA_PATTERN)
    return list(sorted(files, key=lambda fn: int(fn[-6:-4])))


def get_reference_array():
    files = get_reference_files()
    arrs = [
        np.asarray(Image.open(fn))
        for fn in files
    ]
    return np.stack(arrs)


@pytest.fixture(scope='module')
def default_tvips_raw():
    return get_reference_array()


@needsdata
def test_comparison(default_tvips, default_tvips_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_tvips_raw, (-1, *tuple(default_tvips.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_tvips)


@needsdata
def test_comparison_roi(default_tvips, default_tvips_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_tvips.shape.nav),
        p=[0.1, 0.9]
    )
    udf = ValidationUDF(reference=default_tvips_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_tvips, roi=roi)


@pytest.fixture
def default_tvips(lt_ctx):
    nav_shape = (10,)
    ds = lt_ctx.load(
        "tvips",
        path=TVIPS_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=MMapBackend(),
    )
    ds.set_num_cores(4)
    return ds


@pytest.fixture
def buffered_tvips(lt_ctx):
    nav_shape = (10,)
    ds = lt_ctx.load(
        "tvips",
        path=TVIPS_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=BufferedBackend(),
    )
    return ds


@pytest.fixture
def direct_tvips(lt_ctx):
    nav_shape = (10,)
    ds = lt_ctx.load(
        "tvips",
        path=TVIPS_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=DirectBackend(),
    )
    return ds


def test_glob(tmp_path, naughty_filename):
    naughty_dir = tmp_path / naughty_filename
    naughty_dir.mkdir()

    all_naughty = []
    for k in (0, 1, 2):
        naughty = naughty_dir / (naughty_filename + f'_{k:03}.tvips')
        print(naughty)
        naughty.touch()
        all_naughty.append(naughty)

    filenames_1 = {
        Path(f)
        for f in get_filenames(naughty)
    }
    print(filenames_1)
    target = set(all_naughty)
    assert filenames_1 == target


@needsdata
def test_detect(lt_ctx):
    params = TVIPSDataSet.detect_params(TVIPS_TESTDATA_PATH, lt_ctx.executor)["parameters"]
    assert params == {
        "path": TVIPS_TESTDATA_PATH,
        "nav_shape": (10,),
        "sig_shape": (512, 512),
        "sync_offset": 0,
    }


@needsdata
def test_positive_sync_offset(default_tvips, lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = TVIPSDataSet(
        path=TVIPS_TESTDATA_PATH, sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(2)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (3,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (5, 0, 0)
    assert p.slice.shape[0] == 5

    result = lt_ctx.run_udf(dataset=default_tvips, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


@needsdata
def test_negative_sync_offset(default_tvips, lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = TVIPSDataSet(
        path=TVIPS_TESTDATA_PATH, sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(2)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (3,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (5, 0, 0)
    assert p.slice.shape[0] == 5

    result = lt_ctx.run_udf(dataset=default_tvips, udf=udf)
    result = result['intensity'].raw_data[:default_tvips._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@needsdata
def test_offset_smaller_than_nav_shape(lt_ctx):
    nav_shape = (10,)
    sync_offset = -1030

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "tvips",
            path=TVIPS_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10, 10\), which is \(-image_count, image_count\)"
        )


@needsdata
def test_offset_greater_than_nav_shape(lt_ctx):
    nav_shape = (10,)
    sync_offset = 1030

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "tvips",
            path=TVIPS_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-10, 10\), which is \(-image_count, image_count\)"
        )


@needsdata
@pytest.mark.with_numba
def test_read(default_tvips):
    partitions = default_tvips.get_partitions()
    p = next(partitions)
    assert len(p.shape) == 3
    assert tuple(p.shape[1:]) == (512, 512)

    tileshape = Shape(
        (2,) + tuple(default_tvips.shape.sig),
        sig_dims=default_tvips.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_tvips.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == (2, 512, 512)
    for _ in tiles:
        pass


@needsdata
def test_scheme_too_large(default_tvips):
    partitions = default_tvips.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_tvips.shape.sig),
        sig_dims=default_tvips.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_tvips.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_tvips.shape.sig)
    for _ in tiles:
        pass


@needsdata
def test_pickle_is_small(default_tvips):
    pickled = cloudpickle.dumps(default_tvips)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 2 * 1024


@needsdata
@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction(default_tvips, lt_ctx, with_roi):
    ds = default_tvips

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[1] = True
        roi[3] = True
        roi[7] = True
    else:
        roi = None
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx, exclude=[(45, 144), (124, 30)])
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@needsdata
@pytest.mark.with_numba
def test_with_roi(default_tvips, lt_ctx):
    udf = PickUDF()
    roi = np.zeros(default_tvips.shape.nav, dtype=bool)
    roi[0] = 1
    res = lt_ctx.run_udf(udf=udf, dataset=default_tvips, roi=roi)
    np.array(res['intensity']).shape == (1, 512, 512)


@needsdata
def test_diagnostics(default_tvips):
    print(default_tvips.diagnostics)


@needsdata
@pytest.mark.dist
def test_tvips_dist(dist_ctx):
    nav_shape = (10,)
    ds = TVIPSDataSet(path=TVIPS_TESTDATA_PATH, nav_shape=nav_shape)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    results = dist_ctx.run(analysis)
    assert results[0].raw_data.shape == (512, 512)


@needsdata
def test_reshape_nav(lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("tvips", path=TVIPS_TESTDATA_PATH, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("tvips", path=TVIPS_TESTDATA_PATH, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("tvips", path=TVIPS_TESTDATA_PATH, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


@needsdata
def test_incorrect_sig_shape(lt_ctx):
    nav_shape = (10,)
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "tvips",
            path=TVIPS_TESTDATA_PATH,
            nav_shape=nav_shape,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 262144"
    )


@needsdata
def test_compare_backends(lt_ctx, default_tvips, buffered_tvips):
    x = random.choice(range(default_tvips.shape.nav[0]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_tvips,
        x=x,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_tvips,
        x=x,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_tvips, direct_tvips):
    x = random.choice(range(default_tvips.shape.nav[0]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_tvips,
        x=x,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_tvips,
        x=x,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_tvips, buffered_tvips, as_sparse):
    roi = np.zeros(default_tvips.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[6] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_tvips, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_tvips, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("tvips", TVIPS_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



import logging

import numba
import scipy.sparse
import numpy as np

from .cache import cached_njit  # NOQA: F401

logger = logging.getLogger(__name__)

numba_dtypes = frozenset(np.dtype(t) for t in (
    bool,
    int,
    float,
    np.bool_,
    np.uint8,
    np.int8,
    np.uint16,
    np.int16,
    np.uint32,
    np.int32,
    np.uint64,
    np.int64,
    np.float32,
    np.float64,
    np.complex64,
    np.complex128,
    np.intc,
    np.intp,
    np.uintc,
    np.uintp
))


@numba.njit(boundscheck=True, nogil=True)
def numba_ravel_multi_index_single(multi_index, dims):
    # only supports the "single index" case
    idxs = range(len(dims) - 1, -1, -1)
    res = 0
    for idx in idxs:
        stride = 1
        for dimidx in range(idx + 1, len(dims)):
            stride *= dims[dimidx]
        res += multi_index[idx] * stride
    return res


@numba.njit(boundscheck=True, nogil=True)
def numba_ravel_multi_index_multi(multi_index, dims):
    # only supports the "multi index" case
    idxs = range(len(dims) - 1, -1, -1)
    res = np.zeros(len(multi_index[0]), dtype=np.intp)
    for i in range(len(res)):
        for idx in idxs:
            stride = 1
            for dimidx in range(idx + 1, len(dims)):
                stride *= dims[dimidx]
            res[i] += multi_index[idx, i] * stride
    return res


@numba.njit(boundscheck=True, nogil=True)
def numba_unravel_index_single(index, shape):
    sizes = np.zeros(len(shape), dtype=np.intp)
    result = np.zeros(len(shape), dtype=np.intp)
    sizes[-1] = 1
    for i in range(len(shape) - 2, -1, -1):
        sizes[i] = sizes[i + 1] * shape[i + 1]
    remainder = index
    for i in range(len(shape)):
        result[i] = remainder // sizes[i]
        remainder %= sizes[i]
    return result


@numba.njit(boundscheck=True, nogil=True)
def numba_unravel_index_multi(indices, shape):
    sizes = np.zeros(len(shape), dtype=np.intp)
    result = np.zeros((len(shape), len(indices)), dtype=np.intp)
    sizes[-1] = 1
    for i in range(len(shape) - 2, -1, -1):
        sizes[i] = sizes[i + 1] * shape[i + 1]
    remainder = indices
    for i in range(len(shape)):
        result[i] = remainder // sizes[i]
        remainder %= sizes[i]
    return result


def rmatmul(left_dense, right_sparse):
    '''
    Custom implementations for dense-sparse matrix product

    Currently the implementation of __rmatmul__ in scipy.sparse
    uses transposes and the left hand side matrix product,
    which leads to poor performance for large tiles.

    See https://github.com/scipy/scipy/issues/13211
    See https://github.com/LiberTEM/LiberTEM/issues/917

    Parameters
    ----------

    left_dense : numpy.ndarray
        2D left-hand matrix (dense)

    right_sparse : Union[scipy.sparse.csc_matrix, scipy.sparse.csr_matrix]
        2D right hand sparse matrix

    Returns
    -------
    numpy.ndarray : Result of matrix product

    '''
    if len(left_dense.shape) != 2:
        raise ValueError(f"Shape of left_dense is not 2D, but {left_dense.shape}.")
    if len(right_sparse.shape) != 2:
        raise ValueError(f"Shape of right_sparse is not 2D, but {right_sparse.shape}.")
    if left_dense.shape[1] != right_sparse.shape[0]:
        raise ValueError(
            f"Shape mismatch: left_dense.shape[1] != right_sparse.shape[0], "
            f"got {left_dense.shape[1], right_sparse.shape[0]}."
        )
    result_t = np.zeros(
        shape=(right_sparse.shape[1], left_dense.shape[0]),
        dtype=np.result_type(right_sparse, left_dense)
    )

    if isinstance(right_sparse, scipy.sparse.csc_matrix):
        _rmatmul_csc(
            left_dense=left_dense,
            right_data=right_sparse.data,
            right_indices=right_sparse.indices,
            right_indptr=right_sparse.indptr,
            res_inout_t=result_t
        )
    elif isinstance(right_sparse, scipy.sparse.csr_matrix):
        _rmatmul_csr(
            left_dense=left_dense,
            right_data=right_sparse.data,
            right_indices=right_sparse.indices,
            right_indptr=right_sparse.indptr,
            res_inout_t=result_t
        )
    else:
        raise ValueError(
            f"Right hand matrix mus be of type scipy.sparse.csc_matrix or scipy.sparse.csr_matrix, "
            f"got {type(right_sparse)}."
        )
    return result_t.T.copy()


@numba.njit(fastmath=True, cache=True, nogil=True)
def _rmatmul_csc(left_dense, right_data, right_indices, right_indptr, res_inout_t):
    left_rows = left_dense.shape[0]
    for right_column in range(len(right_indptr) - 1):
        offset = right_indptr[right_column]
        items = right_indptr[right_column+1] - offset
        if items > 0:
            for i in range(items):
                index = i + offset
                right_row = right_indices[index]
                right_value = right_data[index]
                for left_row in range(left_rows):
                    tmp = left_dense[left_row, right_row] * right_value
                    res_inout_t[right_column, left_row] += tmp


@numba.njit(fastmath=True, cache=True, nogil=True)
def _rmatmul_csr(left_dense, right_data, right_indices, right_indptr, res_inout_t):
    left_rows = left_dense.shape[0]
    rowbuf = np.empty(shape=(left_rows,), dtype=left_dense.dtype)
    for right_row in range(len(right_indptr) - 1):
        offset = right_indptr[right_row]
        items = right_indptr[right_row+1] - offset
        if items > 0:
            rowbuf[:] = left_dense[:, right_row]
            for i in range(items):
                index = i + offset
                right_column = right_indices[index]
                right_value = right_data[index]
                for left_row in range(left_rows):
                    tmp = rowbuf[left_row] * right_value
                    res_inout_t[right_column, left_row] += tmp



import os
import uuid
import pickle

import pytest

from libertem.io.fs import _get_alt_path, FSError


@pytest.mark.skipif(os.name != 'nt',
                    reason="only runs on windows")
def test_doesnt_exist_windows():
    rnd = str(uuid.uuid4())
    path = r"C:\\" + rnd + r"\really\this\does\not\exist\\"

    alt = str(_get_alt_path(path))
    assert alt == "C:\\"


@pytest.mark.skipif(os.name == 'nt',
                    reason="doesnt run on windows")
def test_doesnt_exist_posix():
    rnd = str(uuid.uuid4())
    path = "/%s/really/this/doesnt/exist/" % rnd

    alt = str(_get_alt_path(path))
    assert alt == "/"


def test_pickle_fserror():
    e = FSError("wat", code="WAT", alternative="/wat/")
    pickled = pickle.dumps(e)
    pickle.loads(pickled)



import numpy as np
import numba

from libertem.common.numba import cached_njit


def fn(a, b):
    return a ** b


cached_fn = cached_njit(fastmath=True)(fn)
numba_fn = numba.njit(fastmath=True, cache=True)(fn)

cached_fn(np.float32(1.2345), 7)  # call 1
numba_fn(np.float32(1.2345), 7)  # call 2



from typing import Any, Optional, TYPE_CHECKING
from collections.abc import Sequence

import jsonschema
import numpy as np
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common import Shape

if TYPE_CHECKING:
    from numpy import typing as nt


class DataSetMeta:
    """
    shape
        "native" dataset shape, can have any dimensionality

    array_backends: Optional[Sequence[ArrayBackend]]

    raw_dtype : np.dtype
        dtype used internally in the data set for reading

    dtype : np.dtype
        Best-fitting output dtype. This can be different from raw_dtype, for example
        if there are post-processing steps done as part of reading, which need a different
        dtype. Assumed equal to raw_dtype if not given

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    image_count
        Total number of frames in the dataset

    metadata
        Any metadata offered by the DataSet, not specified yet
    """
    def __init__(
        self,
        shape: Shape,
        array_backends: Optional[Sequence[ArrayBackend]] = None,
        image_count: int = 0,
        raw_dtype: "Optional[nt.DTypeLike]" = None,
        dtype: "Optional[nt.DTypeLike]" = None,
        metadata: Optional[Any] = None,
        sync_offset: int = 0
    ):
        self.shape = shape
        if array_backends is None:
            array_backends = (NUMPY, CUDA)
        self.array_backends = array_backends
        if dtype is None:
            dtype = raw_dtype
        self.dtype: np.dtype = np.dtype(dtype)
        self.raw_dtype: np.dtype = np.dtype(raw_dtype)
        self.image_count = image_count
        self.sync_offset = sync_offset
        self.metadata = metadata

    def __getitem__(self, key):
        return self.metadata[key]


class PartitionStructure:
    """
    Structure of the dataset.

    Assumed to be contiguous on the flattened navigation axis.

    Parameters
    ----------

    slices : List[Tuple[Int, ...]]
        List of tuples [start_idx, end_idx) that partition the data set by the flattened
        navigation axis

    shape : Shape
        shape of the whole dataset

    dtype : numpy dtype
        The dtype of the data as it is on disk. Can contain endian indicator, for
        example >u2 for big-endian 16bit data.
    """
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/PartitionStructure.schema.json",
        "title": "PartitionStructure",
        "type": "object",
        "properties": {
            "version": {"const": 1},
            "slices": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": {
                        "type": "number", "minItems": 2, "maxItems": 2,
                    }
                },
                "minItems": 1,
            },
            "shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
            },
            "sig_dims": {"type": "number"},
            "dtype": {"type": "string"},
        },
        "required": ["version", "slices", "shape", "sig_dims", "dtype"]
    }

    def __init__(self, shape, slices, dtype):
        self.slices = slices
        self.shape = shape
        self.dtype = np.dtype(dtype)

    def serialize(self):
        data = {
            "version": 1,
            "slices": [[s[0], s[1]] for s in self.slices],
            "shape": list(self.shape),
            "sig_dims": self.shape.sig.dims,
            "dtype": str(self.dtype),
        }
        jsonschema.validate(schema=self.SCHEMA, instance=data)
        return data

    @classmethod
    def from_json(cls, data):
        jsonschema.validate(schema=cls.SCHEMA, instance=data)
        shape = Shape(tuple(data["shape"]), sig_dims=data["sig_dims"])
        return PartitionStructure(
            slices=[tuple(item) for item in data["slices"]],
            shape=shape,
            dtype=np.dtype(data["dtype"]),
        )

    @classmethod
    def from_ds(cls, ds):
        data = {
            "version": 1,
            "slices": [
                [p.slice.origin[0], p.slice.origin[0] + p.slice.shape[0]]
                for p in ds.get_partitions()
            ],
            "shape": list(ds.shape),
            "sig_dims": ds.shape.sig.dims,
            "dtype": str(ds.dtype),
        }
        return cls.from_json(data)



from collections import defaultdict
import functools

import numpy as np
import numba

from libertem.udf import UDF
from libertem.common.buffers import reshaped_view


@numba.njit(fastmath=True, cache=True, nogil=True)
def merge_single(n, n_0, sum_0, varsum_0, n_1, sum_1, varsum_1, mean_1):
    '''
    Basic function to perform numerically stable merge.

    This function is designed to be inlined in a loop over all pixels in a frame
    to merge an individual pixel, with some parts pre-calculated.

    Based on :cite:`Schubert2018`.

    Parameters
    ----------
    n : int
        Total number of frames, assumed n == n_0 + n_1
        Pre-calculated since equal for all pixels.
    n_0 : int
        Number of frames aggregated in sum_0 and varsum_0, n_0 > 0.
        The case n_0 == 0 is handled separately in the calling function.
    sum_0, varsum_0 : float
        Aggregate sum and sum of variances
    n_1 : int
        Number of frames to merge from minibatch, n_1 > 0
    sum_1, varsum_1, mean_1 : float
        Minibatch sum, sum of variances and mean. The mean is
        available from the previous minibatch calculation in the innermost
        aggregation loop as a "hot" value.samples

    Returns
    -------
    sum, varsum : float
        New aggregate sum and sum of variances
    '''
    # compute mean for each partitions
    mean_0 = sum_0 / n_0

    # compute mean for joint samples
    delta = mean_1 - mean_0
    mean = mean_0 + (n_1 * delta) / n

    # compute sum of images for joint samples
    sumsum = sum_0 + sum_1

    # compute sum of variances for joint samples
    partial_delta = mean_1 - mean
    # The original version had delta * partial_delta, which gave wrong results
    # for complex numbers. Taking the absolute seems to fix this for complex.
    # For real values it is equivalent since delta and partial_delta always have
    # the same sign: inserting mean from above yields
    # partial_delta = delta + (n_1 * delta) / n
    # n_1 and n are natural numbers.
    # For complex values it is plausible since the variance
    # sum of the total is a real number. If the mean of both subsets is equal,
    # the total is varsum_0 + varsum_1, since both delta and partial_delta are
    # zero. If the mean of the subsets is different, the total variance sum
    # must always be larger than the variance sums of the
    # subsets. Taking the absolute of both is the only plausible way I can see
    # to make sure the calculation is equivalent for real values and at the same time always
    # a positive real value for complex numbers.
    # FIXME proper validation and proof for complex
    varsum = varsum_0 + varsum_1 + (n_1 * np.abs(delta) * np.abs(partial_delta))
    return sumsum, varsum


@numba.njit(nogil=True)
def merge(dest_n, dest_sum, dest_varsum, src_n, src_sum, src_varsum, src_mean):
    """
    Given two sets of buffers, with sum of frames
    and sum of variances, aggregate joint sum of frames
    and sum of variances in destination buffers using one pass
    algorithm.

    This is ther numerical workhorse for :meth:`StdDevUDF.merge`.

    Based on :cite:`Schubert2018`.

    Parameters
    ----------
    dest_n : int
        Number of frames aggregated in dest_sum and dest_varsum
        The case :code:`dest_N == 0` is handled correctly.
    dest_sum, dest_varsum : one-dimensional numpy.ndarray
        Aggregation buffers, will be updated: sum of pixels, sum of variances
    src_n : int
        Number of frames aggregated in src_sum and src_varsum
    src_sum, src_varsum : one-dimensional numpy.ndarray
        Source buffers to merge: sum of pixels, sum of variances


    Returns
    -------
    N:
        New number of frames aggregated in aggregation buffer
    """
    if dest_n == 0:
        dest_sum[:] = src_sum
        dest_varsum[:] = src_varsum
        return src_n
    elif src_n == 0:
        # Can happen from empty partitions due to sync offset
        return dest_n
    else:
        n = dest_n + src_n

        for pixel in range(len(dest_sum)):
            dest_sum[pixel], dest_varsum[pixel] = merge_single(
                n,
                dest_n, dest_sum[pixel], dest_varsum[pixel],
                src_n, src_sum[pixel], src_varsum[pixel], src_mean[pixel]
            )
    return n


@numba.njit(fastmath=True, nogil=True)
def process_tile(tile, n_0, sum_inout, varsum_inout):
    '''
    Compute sum and variance of :code:`tile` along navigation axis
    and merge into aggregation buffers. Numerical "workhorse" for
    :meth:`StdDevUDF.process_tile`.

    Parameters
    ----------
    tile : 2-dimensional numpy.ndarray
        Tile with flattened signal dimension
    n_0 : int
        Number of frames already aggegated in sum_inout, varsum_inout
        Cannot be 0 -- the initial case is handled in :meth:`StdDevUDF.process_tile`
        For unknown reasons, handling the case N0 == 0 in this function degraded performance
        significantly. Re-check with a newer compiler version!
    sum_inout, varsum_inout : numpy.ndarray
        Aggregation buffers (1D) with length matching the flattened signal dimension
        of the tile. The tile will be merged into these buffers (call by reference)

    Returns
    -------
    n : int
        New number of frames in aggregation buffer
    '''
    n_frames = tile.shape[0]
    n_pixels = tile.shape[1]
    n = n_0 + n_frames

    BLOCKSIZE = 256
    n_blocks = n_pixels // BLOCKSIZE

    sumsum = np.zeros(BLOCKSIZE, dtype=sum_inout.dtype)
    varsum = np.zeros(BLOCKSIZE, dtype=varsum_inout.dtype)

    for block in range(n_blocks):
        pixel_offset = block*BLOCKSIZE
        sumsum[:] = 0
        for frame in range(n_frames):
            for i in range(BLOCKSIZE):
                sumsum[i] += tile[frame, pixel_offset + i]
        mean = sumsum / n_frames
        varsum[:] = 0
        for frame in range(n_frames):
            for i in range(BLOCKSIZE):
                varsum[i] += np.abs(tile[frame, pixel_offset + i] - mean[i])**2
        for i in range(BLOCKSIZE):
            sum_inout[pixel_offset + i], varsum_inout[pixel_offset + i] = merge_single(
                n,
                n_0, sum_inout[pixel_offset + i], varsum_inout[pixel_offset + i],
                n_frames, sumsum[i], varsum[i], mean[i]
            )
    for pixel in range(n_blocks*BLOCKSIZE, n_pixels):
        sumsum_rest = 0.
        for frame in range(n_frames):
            sumsum_rest += tile[frame, pixel]
        mean_rest = sumsum_rest / n_frames
        varsum_rest = 0.
        for frame in range(n_frames):
            varsum_rest += np.abs(tile[frame, pixel] - mean_rest)**2
        sum_inout[pixel], varsum_inout[pixel] = merge_single(
            n,
            n_0, sum_inout[pixel], varsum_inout[pixel],
            n_frames, sumsum_rest, varsum_rest, mean_rest
        )
    return n


def merge_ndarray(dest_n, dest_sum, dest_varsum, src_n, src_sum, src_varsum, src_mean, forbuf):
    '''
    Version of :func:`merge` that only uses ndarray routines
    for CuPy and sparse input data
    '''
    if dest_n == 0:
        dest_sum[:] = forbuf(src_sum, dest_sum)
        dest_varsum[:] = forbuf(src_varsum, dest_varsum)
        return src_n
    elif src_n == 0:
        return dest_n
    else:
        n = dest_n + src_n
        # compute mean for each partitions
        mean_0 = dest_sum / dest_n
        mean_1 = src_mean

        # compute mean for joint samples
        delta = mean_1 - mean_0
        mean = mean_0 + (src_n * delta) / n

        # compute sum of images for joint samples
        dest_sum += forbuf(src_sum, dest_sum)

        # compute sum of variances for joint samples
        partial_delta = mean_1 - mean
        dest_varsum += forbuf(
            # Use multiply for compatibility with matrix interface (scipy.sparse)
            src_varsum + (np.multiply(np.multiply(src_n, np.abs(delta)), np.abs(partial_delta))),
            dest_varsum
        )
        return n


def process_tile_ndarray(tile, n_0, sum_inout, varsum_inout, forbuf):
    '''
    Version of :func:`process_tile` that only uses ndarray routines
    for CuPy and sparse input data
    '''
    n_frames = tile.shape[0]
    n = n_0 + n_frames
    sumsum = np.sum(tile, axis=0)
    mean = sumsum / n_frames
    delta = np.abs(tile - mean)
    varsum = np.sum(np.multiply(delta, delta), axis=0)
    merge_ndarray(
        n_0, sum_inout, varsum_inout,
        n_frames, sumsum, varsum, mean,
        forbuf
    )
    return n


# Helper function to make sure the frame count
# is consistent at the merge stage
def _validate_n(num_frame):
    if len(num_frame) == 0:
        return 0
    else:
        values = tuple(num_frame.values())
        assert np.all(np.equal(values, values[0]))
        return values[0]


class StdDevUDF(UDF):
    """
    Compute sum of variances and sum of pixels from the given dataset

    The one-pass algorithm used in this code is taken from the following
    paper: :cite:`Schubert2018`.

    .. versionchanged:: 0.5.0
        Result buffers have been renamed

    .. versionchanged:: 0.7.0
        :code:`var`, :code:`mean`, and :code:`std` are now returned directly
        from the UDF via :code:`get_results`.

    .. versionchanged:: 0.11
        added dtype and use_numba parameters, added support for sparse input.

    Parameters
    ----------

    dtype
        Base dtype to determine numerical precision.

    use_numba : bool
        Use the Numba-accelerated version if input array format and backend allow.

    Examples
    --------

    >>> udf = StdDevUDF()
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> np.array(result["varsum"])        # variance times number of frames
    array(...)
    >>> np.array(result["num_frames"])  # number of frames for each tile
    array(...)
    >>> np.array(result["sum"])  # sum of all frames
    array(...)
    >>> np.array(result["var"])
    array(...)
    >>> np.array(result["mean"])
    array(...)
    >>> np.array(result["std"])
    array(...)
    """
    def __init__(self, dtype=None, use_numba=True) -> None:
        super().__init__(dtype=dtype, use_numba=use_numba)

    def get_backends(self):
        # TODO supporting sparse backends efficiently requires a dedicated implementation
        # since the generic one calculates the difference from a mean value,
        # which can densify a sparse array.
        return tuple(b for b in self.BACKEND_ALL if b not in self.SPARSE_BACKENDS)

    def get_result_buffers(self):
        ''
        base_dtype = self.params.dtype
        if base_dtype is None:
            base_dtype = np.float64
        dtype = np.result_type(self.meta.input_dtype, base_dtype)
        return {
            'varsum': self.buffer(
                kind='sig', dtype=base_dtype, where='device'
            ),
            'num_frames': self.buffer(
                kind='single', dtype='int64'
            ),
            'sum': self.buffer(
                kind='sig', dtype=dtype, where='device'
            ),
            'var': self.buffer(
                kind='sig', dtype=base_dtype, use='result_only',
            ),
            'std': self.buffer(
                kind='sig', dtype=base_dtype, use='result_only',
            ),
            'mean': self.buffer(
                kind='sig', dtype=dtype, use='result_only',
            ),
        }

    def get_task_data(self):
        return {
            'num_frames': defaultdict(int)
        }

    def postprocess(self):
        self.results.num_frames[:] = _validate_n(self.task_data.num_frames)

    def merge(self, dest, src):
        ''
        # Given destination and source buffers that contain sum of variances, sum of frames,
        # and the number of frames used in each of the buffers, merge the source
        # buffers into the destination buffers by computing the joint sum of variances and
        # sum of frames over all frames used

        # Parameters
        # ----------
        # dest
        #     Aggregation bufer that contains sum of variances, sum of frames, and the
        #     number of frames
        # src
        #     Partial results that contains sum of variances, sum of frames, and the
        #     number of frames of a partition to be merged into the aggregation buffers
        dest_n = dest.num_frames[0]
        src_n = src.num_frames[0]

        if self.params.use_numba:
            my_merge = merge
        else:
            my_merge = functools.partial(merge_ndarray, forbuf=self.forbuf)

        n = my_merge(
            dest_n=dest_n,
            dest_sum=reshaped_view(dest.sum, (-1,)),
            dest_varsum=reshaped_view(dest.varsum, (-1,)),
            src_n=src_n,
            src_sum=reshaped_view(src.sum, (-1,)),
            src_varsum=reshaped_view(src.varsum, (-1,)),
            src_mean=reshaped_view(src.sum, (-1,)) / src_n,
        )
        dest.num_frames[:] = n

    def merge_all(self, ordered_results):
        ''
        n_frames = np.stack([b.num_frames[0] for b in ordered_results.values()])
        pixel_sums = np.stack([b.sum for b in ordered_results.values()])
        pixel_varsums = np.stack([b.varsum for b in ordered_results.values()])

        # Expand n_frames to be broadcastable
        extra_dims = pixel_sums.ndim - n_frames.ndim
        n_frames = n_frames.reshape(n_frames.shape + (1,) * extra_dims)

        cumulative_frames = np.cumsum(n_frames, axis=0)
        cumulative_sum = np.cumsum(pixel_sums, axis=0)
        sumsum = cumulative_sum[-1, ...]
        total_frames = cumulative_frames[-1, 0]

        mean_0 = cumulative_sum / cumulative_frames
        # Handle the fact that mean_0 is indexed to results from
        # up-to the partition before. We shift everything one to
        # the right, and we don't care about result 0 because it
        # is by definiition replaced with varsum[0, ...]
        mean_0 = np.roll(mean_0, 1, axis=0)

        mean_1 = pixel_sums / n_frames
        delta = mean_1 - mean_0
        mean = mean_0 + (n_frames * delta) / cumulative_frames
        partial_delta = mean_1 - mean
        varsum = pixel_varsums + (n_frames * np.abs(delta) * np.abs(partial_delta))
        varsum_total = np.sum(varsum, axis=0)

        return {
            'sum': sumsum,
            'varsum': varsum_total,
            'num_frames': total_frames
        }

    def _adjust_dtype(self, input):
        base_dtype = self.params.dtype
        if base_dtype is None:
            base_dtype = np.float64
        dtype = np.result_type(input.dtype, base_dtype)

        if input.dtype != dtype:
            return input.astype(dtype)
        else:
            return input

    def process_tile(self, tile):
        # Calculate a sum and variance minibatch for the tile and update partition buffers
        # with it.
        key = self.meta.tiling_scheme_idx
        n_0 = self.task_data.num_frames[key]
        n_1 = tile.shape[0]

        if n_0 == 0:
            # Make sure we calculate with full precision
            tile = self._adjust_dtype(tile)
            self.results.sum[:] = self.forbuf(tile.sum(axis=0), self.results.sum)
            # Done this way to support the various array libraries
            delta = np.abs(tile - tile.mean(axis=0))
            self.results.varsum[:] = self.forbuf(
                np.sum(np.multiply(delta, delta), axis=0).real,
                self.results.varsum
            )
            self.task_data.num_frames[key] = n_1
        else:
            if isinstance(tile, np.ndarray) and self.params.use_numba:
                my_process_tile = process_tile
            else:
                # Make sure we calculate with full precision
                tile = self._adjust_dtype(tile)
                my_process_tile = functools.partial(process_tile_ndarray, forbuf=self.forbuf)
            self.task_data.num_frames[key] = my_process_tile(
                tile=tile.reshape((n_1, -1)),
                n_0=n_0,
                sum_inout=reshaped_view(self.results.sum, (-1, )),
                varsum_inout=reshaped_view(self.results.varsum, (-1, )),
            )

    def get_results(self):
        ''
        # Calculate variance, mean and standard deviation from raw UDF results
        num_frames = self.results.num_frames[0]

        var = self.results.varsum / num_frames

        return {
            'var': var,
            'std': np.sqrt(var),
            'mean': self.results.sum / num_frames,
        }


def consolidate_result(udf_result):
    '''
    Calculate variance, mean and standard deviation
    from raw UDF results and consolidate the per-tile frame counter
    into a single value. Convert all result arrays to `ndarray`.

    Note
    ----
    This is mostly here for backwards-compatability - nowadays, 'var', 'std',
    and 'mean' are already calculated in :meth:`StdDevUDF.get_results`.

    Parameters
    ----------
    udf_result : Dict[str, BufferWrapper]
        UDF result with keys 'sum', 'varsum', 'num_frames', 'var', 'std', 'mean'

    Returns
    -------
    pass_results : Dict[str, Union[numpy.ndarray, int]]
        Result dictionary with keys :code:`'sum', 'varsum', 'var', 'std', 'mean'` as
        :class:`numpy.ndarray`, and :code:`'num_frames'` as :code:`int`
    '''
    return {
        'num_frames': udf_result['num_frames'].data[0],
        'varsum': udf_result['varsum'].data,
        'sum': udf_result['sum'].data,
        'var': udf_result['var'].data,
        'std': udf_result['std'].data,
        'mean': udf_result['mean'].data,
    }


def run_stddev(ctx, dataset, roi=None, progress=False, use_numba=True):
    """
    Compute sum of variances and sum of pixels from the given dataset

    One-pass algorithm used in this code is taken from the following paper:
    :cite:`Schubert2018`.

    .. versionchanged:: 0.5.0
        Result buffers have been renamed

    .. versionchanged:: 0.5.0
        Added :code:`progress` parameter for progress bar

    Parameters
    ----------
    ctx : libertem.api.Context
    dataset : libertem.io.dataset.base.DataSet
        dataset to work on
    roi : numpy.ndarray
        Region of interest, see :ref:`udf roi` for more information.
    progress : bool, optional
        Show progress bar. Default is :code:`False`.
    use_numba : bool
        Use the Numba backend if available, otherwise use ndarray based implementation

    Returns
    -------
    pass_results
        A dictionary of narrays that contains sum of variances, sum of pixels,
        and number of frames used to compute the above statistic

    To retrieve statistic, using the following commands:
    variance : :code:`pass_results['var']`
    standard deviation : :code:`pass_results['std']`
    sum of pixels : :code:`pass_results['sum']`
    mean : :code:`pass_results['mean']`
    number of frames : :code:`pass_results['num_frames']`
    """
    stddev_udf = StdDevUDF(use_numba=use_numba)
    pass_results = ctx.run_udf(
        dataset=dataset, udf=stddev_udf, roi=roi, progress=progress
    )

    return consolidate_result(pass_results)



import pytest


from libertem.udf.base import get_resources_for_backends


def test_no_common_backends():
    with pytest.raises(ValueError):
        get_resources_for_backends(
            [('numpy',), ('cupy',)], user_backends=None,
        )


def test_no_common_backends_with_user():
    with pytest.raises(ValueError):
        get_resources_for_backends(
            [('numpy', 'cupy'), ('cupy',)], user_backends=('numpy',),
        )


def test_common_backends_cpu_1():
    resources = get_resources_for_backends(
        [('numpy', 'cupy'), ('numpy',)], user_backends=None,
    )
    assert resources == {
        'CPU': 1, 'compute': 1, 'ndarray': 1
    }


def test_common_backends_cpu_2():
    resources = get_resources_for_backends(
        [('numpy', 'cupy'), ('numpy', 'cupy',)], user_backends=('numpy',),
    )
    assert resources == {
        'CPU': 1, 'compute': 1, 'ndarray': 1
    }


def test_common_backends_compute_1():
    resources = get_resources_for_backends(
        [('numpy', 'cupy'), ('numpy', 'cupy',)], user_backends=None,
    )
    assert resources == {
        'compute': 1, 'ndarray': 1
    }


def test_common_backends_compute_2():
    resources = get_resources_for_backends(
        [('numpy', 'cupy'), ('numpy', 'cupy',)], user_backends=('numpy', 'cupy'),
    )
    assert resources == {
        'compute': 1, 'ndarray': 1
    }


def test_common_backends_gpu_1():
    resources = get_resources_for_backends(
        [('cupy',), ('cuda',)], user_backends=None,
    )
    assert resources == {
        'CUDA': 1, 'compute': 1, 'ndarray': 1
    }


def test_common_backends_gpu_2():
    resources = get_resources_for_backends(
        [('cuda',), ('cupy',)], user_backends=('cupy', 'cuda'),
    )
    assert resources == {
        'CUDA': 1, 'compute': 1, 'ndarray': 1
    }


def test_common_backends_gpu_3():
    resources = get_resources_for_backends(
        [('cupy',), ('cupy',)], user_backends=('cupy',),
    )
    assert resources == {
        'CUDA': 1, 'compute': 1, 'ndarray': 1
    }


def test_common_backends_as_string():
    resources = get_resources_for_backends(
        ['cupy', ('cupy',)], user_backends='cupy',
    )
    assert resources == {
        'CUDA': 1, 'compute': 1, 'ndarray': 1
    }



from libertem import masks


def get_roi(params, shape):

    if "roi" not in params or ("shape" not in params["roi"]):
        return None
    params = params["roi"]
    ny, nx = tuple(shape)
    if params["shape"] == "disk":
        roi = masks.circular(
            params["cx"],
            params["cy"],
            nx, ny,
            params["r"],
        )
    elif params["shape"] == "rect":
        roi = masks.rectangular(
            params["x"],
            params["y"],
            params["width"],
            params["height"],
            nx, ny,
        )
    else:
        raise NotImplementedError("unknown shape %s" % params["shape"])
    return roi



import os
import json

import pytest
import numpy as np

from libertem.io.dataset.base import PartitionStructure, DataSetException
from libertem.io.dataset.cluster import ClusterDataSet

# FIXME test on actual data structure, including correction
# from utils import dataset_correction_verification


@pytest.fixture(scope='function')
def draw_directory(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('draw_directory')
    yield datadir


def test_structure_from_json():
    data = {
        "version": 1,
        "slices": [[0, 127], [128, 256]],
        "shape": [16, 16, 128, 128],
        "sig_dims": 2,
        "dtype": "float32",
    }
    struct = PartitionStructure.from_json(data)
    assert struct.shape.sig.dims == 2
    assert tuple(struct.shape) == (16, 16, 128, 128)
    assert struct.slices == [(0, 127), (128, 256)]
    assert struct.dtype == np.dtype("float32")


def test_initialization(draw_directory, lt_ctx):
    data = {
        "version": 1,
        "slices": [[0, 8], [16, 256]],
        "shape": [4, 4, 128, 128],
        "sig_dims": 2,
        "dtype": "float32",
    }
    structure = PartitionStructure.from_json(data)
    ds = ClusterDataSet(path=str(draw_directory), structure=structure)
    ds = ds.initialize(lt_ctx.executor)
    ds.check_valid()

    assert set(os.listdir(draw_directory)) == {"parts", "structure.json"}


@pytest.mark.dist
def test_inconsistent_sidecars_raise_error(dist_ctx, draw_directory):
    # tmpdir_factory doesn't output plain strings,
    # and thus draw_directory was some pytest object. we don't have
    # pytest installed on the worker containers, so it failed spectacularly:
    draw_directory = str(draw_directory)
    # first, create consistent ClusterDataSet:
    data = {
        "version": 1,
        "slices": [[0, 4], [5, 16]],
        "shape": [4, 4, 128, 128],
        "sig_dims": 2,
        "dtype": "float32",
    }
    structure = PartitionStructure.from_json(data)
    ds = ClusterDataSet(path=draw_directory, structure=structure)
    ds = ds.initialize(dist_ctx.executor)
    ds.check_valid()

    # then, scramble the sidecars, creating an inconsistent state:
    data = {
        "worker-1": {
            "version": 1,
            "slices": [[0, 8], [9, 16]],
            "shape": [4, 4, 128, 128],
            "sig_dims": 2,
            "dtype": "int32",
        },
        "worker-2": {
            "version": 1,
            "slices": [[0, 4], [5, 16]],
            "shape": [4, 4, 128, 128],
            "sig_dims": 2,
            "dtype": "float32",
        },
    }

    sidecar_path = os.path.join(draw_directory, 'structure.json')

    def _scramble_sidecars():
        import socket
        sidecar_data = data[socket.gethostname()]
        print(sidecar_path, os.listdir(os.path.dirname(sidecar_path)))
        with open(sidecar_path, "w") as fh:
            json.dump(sidecar_data, fh)
        print(sidecar_path, os.listdir(os.path.dirname(sidecar_path)))

    print(
        list(dist_ctx.executor.run_each_host(_scramble_sidecars))
    )

    print(draw_directory)

    dirlists = dist_ctx.executor.run_each_host(lambda: os.listdir(draw_directory))
    for v in dirlists.values():
        print(set(v))

    sidecars = dist_ctx.executor.run_each_host(lambda: open(sidecar_path).read())
    print(sidecars)

    # try to re-open the dataset, should fail:
    ds = ClusterDataSet(path=draw_directory, structure=structure)

    with pytest.raises(DataSetException) as e:
        ds = ds.initialize(dist_ctx.executor)

    assert e.match('inconsistent sidecars, please inspect')


@pytest.mark.dist
def test_missing_sidecars_are_created(draw_directory, dist_ctx):
    # tmpdir_factory doesn't output plain strings,
    # and thus draw_directory was some pytest object. we don't have
    # pytest installed on the worker containers, so it failed spectacularly:
    draw_directory = str(draw_directory)
    data = {
        "version": 1,
        "slices": [[0, 8], [16, 256]],
        "shape": [4, 4, 128, 128],
        "sig_dims": 2,
        "dtype": "float32",
    }
    structure = PartitionStructure.from_json(data)
    ds = ClusterDataSet(path=draw_directory, structure=structure)
    ds = ds.initialize(dist_ctx.executor)
    ds.check_valid()

    dirlists = dist_ctx.executor.run_each_host(lambda: os.listdir(draw_directory))

    for v in dirlists.values():
        assert set(v) == {"parts", "structure.json"}



import numpy as np
from libertem.analysis.disk import DiskMaskAnalysis
import pytest


@pytest.mark.xfail
def test_compare_hdf5_result_with_raw_result(
    lt_ctx, hdf5_same_data_4d, raw_same_dataset_4d
):
    ds_hdf5 = lt_ctx.load(
        "hdf5", path=hdf5_same_data_4d.filename, ds_path="data",
    )
    analysis_hdf5 = DiskMaskAnalysis(
        dataset=ds_hdf5,
        parameters={"cx": 13, "cy": 13, "r": 13}
    )
    result_hdf5 = lt_ctx.run(analysis_hdf5)
    result_hdf5 = result_hdf5['intensity'].raw_data

    analysis_raw = DiskMaskAnalysis(
        dataset=raw_same_dataset_4d,
        parameters={"cx": 13, "cy": 13, "r": 13}
    )
    result_raw = lt_ctx.run(analysis_raw)
    result_raw = result_raw['intensity'].raw_data

    assert not np.any(np.abs(result_hdf5 - result_raw))



import pytest

import numpy as np

from libertem.io.dataset.memory import MemoryDataSet
from libertem.utils.generate import cbed_frame

from utils import _mk_random


@pytest.fixture
def ds_random():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        tileshape=(1, 16, 16),
        num_partitions=2,
    )
    return dataset


def test_smoke(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random, cx=0, cy=0, ri=0, ro=10, n_bins=2, max_order=23
    )
    lt_ctx.run(analysis)


def test_smoke_small(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random, cx=0, cy=0, ri=0, ro=1, n_bins=1, max_order=23
    )
    results = lt_ctx.run(analysis)
    # access result to actually create result list:
    results['absolute_0_0']


def test_smoke_large(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random, cx=0, cy=0, n_bins=1, max_order=23
    )
    lt_ctx.run(analysis)


def test_smoke_two(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random, cx=0, cy=0, ri=0, ro=2, n_bins=2, max_order=2
    )
    results = lt_ctx.run(analysis)
    # access result to actually create result list:
    results.complex_1_2


def test_smoke_defaults(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random
    )
    lt_ctx.run(analysis)


def test_sparse(ds_random, lt_ctx):
    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds_random, use_sparse=True,
    )
    lt_ctx.run(analysis)


def test_symmetries(lt_ctx):
    (d1, i1, p1) = cbed_frame(all_equal=True, radius=3, indices=np.array([(1, 0)]))
    (d2, i2, p2) = cbed_frame(all_equal=True, radius=3, indices=np.array([(-1, 0)]))
    (d3, i3, p3) = cbed_frame(all_equal=True, radius=3, indices=np.array([(1, 0), (-1, 0)]))
    (d4, i4, p4) = cbed_frame(
        all_equal=True, radius=3, indices=np.array([(1, 0), (-1, 0), (0, 1), (0, -1)])
    )

    data = np.zeros((2, 2, *d1[0].shape))
    data[0, 0] = d1[0]
    data[0, 1] = d2[0]
    data[1, 0] = d3[0]
    data[1, 1] = d4[0]

    ds = MemoryDataSet(data=data)

    r = np.linalg.norm(p2[0] - p1[0]) / 2
    cy, cx = (p2[0] + p1[0]) / 2

    analysis = lt_ctx.create_radial_fourier_analysis(
        dataset=ds, cy=cy, cx=cx, ri=0, ro=r+4, n_bins=2, max_order=8
    )

    results = lt_ctx.run(analysis)

    c_0_0 = results.complex_0_0.raw_data
    c_1_0 = results.complex_1_0.raw_data

    assert np.allclose(np.abs(c_0_0), 0)
    assert np.allclose(np.abs(c_1_0), data.sum(axis=(2, 3)))

    c_0_1 = results.complex_0_1.raw_data
    c_1_1 = results.complex_1_1.raw_data

    assert np.allclose(np.abs(c_0_1), 0)

    assert np.allclose(np.abs(c_1_1[1, 0]), 0)
    assert np.allclose(np.abs(c_1_1[1, 1]), 0)

    assert np.all(np.abs(c_1_1[0, 0]) > 0)
    assert np.all(np.abs(c_1_1[0, 1]) > 0)
    assert np.allclose(np.angle(c_1_1[0, 0]), np.pi/2)
    assert np.allclose(np.angle(c_1_1[0, 1]), -np.pi/2)

    c_0_2 = results.complex_0_2.raw_data
    c_1_2 = results.complex_1_2.raw_data

    assert np.allclose(np.abs(c_0_2), 0)

    # 2-fold suppressed for 4-fold symmetry
    assert np.allclose(np.abs(c_1_2[1, 1]), 0)

    assert np.all(np.abs(c_1_2[0, 0]) > 0)
    assert np.all(np.abs(c_1_2[0, 1]) > 0)
    assert np.all(np.abs(c_1_2[1, 0]) > 0)
    # Discontinuity at this point, can be pi or -pi
    assert np.allclose(np.abs(np.angle(c_1_2[0, 0])), np.pi)
    assert np.allclose(np.abs(np.angle(c_1_2[0, 1])), np.pi)
    assert np.allclose(np.abs(np.angle(c_1_2[1, 0])), np.pi)

    c_0_3 = results.complex_0_3.raw_data
    c_1_3 = results.complex_1_3.raw_data

    assert np.allclose(np.abs(c_0_3), 0)
    # odd harmonics suppressed in 2-fold symmetry
    assert np.allclose(np.abs(c_1_3[1]), 0)

    assert np.all(np.abs(c_1_3[0, 0]) > 0)
    assert np.all(np.abs(c_1_3[0, 1]) > 0)
    assert np.allclose(np.angle(c_1_3[0, 0]), -np.pi/2)
    assert np.allclose(np.angle(c_1_3[0, 1]), np.pi/2)

    c_0_4 = results.complex_0_4.raw_data
    c_1_4 = results.complex_1_4.raw_data

    assert np.allclose(np.abs(c_0_4), 0)

    assert np.all(np.abs(c_1_4) > 0)

    assert np.allclose(np.angle(c_1_4[0, 0]), 0)
    assert np.allclose(np.angle(c_1_4[0, 1]), 0)
    assert np.allclose(np.angle(c_1_4[1, 0]), 0)
    assert np.allclose(np.angle(c_1_4[1, 1]), 0)

    c_0_5 = results.complex_0_5.raw_data
    c_1_5 = results.complex_1_5.raw_data

    assert np.allclose(np.abs(c_0_5), 0)

    # odd harmonics suppressed in 2-fold symmetry
    assert np.allclose(np.abs(c_1_5[1]), 0)

    c_0_7 = results.complex_0_5.raw_data
    c_1_7 = results.complex_1_5.raw_data

    assert np.allclose(np.abs(c_0_7), 0)

    # odd harmonics suppressed in 2-fold symmetry
    assert np.allclose(np.abs(c_1_7[1]), 0)

    c_0_8 = results.complex_0_4.raw_data
    c_1_8 = results.complex_1_4.raw_data

    assert np.allclose(np.abs(c_0_8), 0)

    assert np.all(np.abs(c_1_8) > 0)

    assert np.allclose(np.angle(c_1_8[0, 0]), 0)
    assert np.allclose(np.angle(c_1_8[0, 1]), 0)
    assert np.allclose(np.angle(c_1_8[1, 0]), 0)
    assert np.allclose(np.angle(c_1_8[1, 1]), 0)



import warnings

import numpy as np
from libertem.utils import calc_coords, within_frame


warnings.warn(
    'gridmatching and fullmatch are moved to '
    'libertem_blobfinder.common in libertem-blobfinder>=0.6 '
    'and are slated for removal in LiberTEM 0.16.',
    DeprecationWarning
)


class CorrelationResult:
    """
    Container class for the result of correlation-based refinement of peak
    positions within a frame.
    """
    def __init__(self, centers, refineds=None, peak_values=None, peak_elevations=None):
        if refineds is None:
            refineds = centers
        if peak_values is None:
            peak_values = np.ones(len(centers))
        if peak_elevations is None:
            peak_elevations = np.ones(len(centers))
        assert all(len(centers) == len(other) for other in [refineds, peak_values, peak_elevations])
        self.centers = centers
        self.refineds = refineds
        self.peak_values = peak_values
        self.peak_elevations = peak_elevations

    def __len__(self):
        return len(self.centers)


class PointSelection:
    '''
    Class that represents a subset of a correlation result.

    Attributes
    ----------
    selector : numpy.ndarray
        Boolean mask for all points in the correlation result, :code:`True` indicating
        selected points.
    '''
    def __init__(self, correlation_result: CorrelationResult, selector=None):
        self.correlation_result = correlation_result
        if selector is None:
            self.selector = np.ones(len(correlation_result.centers), dtype=bool)
        else:
            assert len(correlation_result.centers) == len(selector)
            self.selector = selector

    @property
    def centers(self):
        '''
        numpy.ndarray : Integer centers (y, x) of correlation result masked with :attr:`selector`
        '''
        return self.correlation_result.centers[self.selector]

    @property
    def refineds(self):
        '''
        numpy.ndarray : Refined float centers (y, x) of correlation result masked
                        with :attr:`selector`
        '''
        return self.correlation_result.refineds[self.selector]

    @property
    def peak_values(self):
        '''
        numpy.ndarray : Peak heights of correlation result masked with :attr:`selector`
        '''
        return self.correlation_result.peak_values[self.selector]

    @property
    def peak_elevations(self):
        '''
        numpy.ndarray : Peak elevations of correlation result masked with :attr:`selector`
        '''
        return self.correlation_result.peak_elevations[self.selector]

    def __len__(self):
        return np.sum(self.selector)

    def new_selector(self, selector):
        new_selector = np.copy(self.selector)
        new_selector[self.selector] = selector
        return new_selector

    def derive(self, selector=None):
        if selector is None:
            selector = self.selector
        return PointSelection(self.correlation_result, selector)


class Matcher:
    '''
    The main job of the Matcher object is managing the matching parameters
    and making them available for the various matching routines.

    Parameters
    ----------

    tolerance : float
        Position tolerance in px for peaks to be considered matches
    min_weight : float
        Minimum peak elevation of a peak to be considered for matching
    min_match : int
        Minimum number of matching peaks to be considered a match.
    '''
    def __init__(self, tolerance=3, min_weight=0.1, min_match=3):
        self.tolerance = tolerance
        self.min_match = min_match
        self.min_weight = min_weight

    def fastmatch(self, centers, zero, a, b, refineds=None, peak_values=None, peak_elevations=None):
        '''
        This function creates a Match object from correlation_result and approximates
        for zero point and lattice vectors a and b.
        This function is much, much faster than the full match.
        It works well to match a large number of point sets
        that share the same lattice vectors, for example from a
        larger grain or monocrystalline material. It rejects
        random points or other lattices in the CorrelationResult,
        provided they are not on near-integer positions of zero, a, b.

        Parameters
        ----------

        centers : numpy.ndarray
            numpy.ndarray of shape (n, 2) with integer centers (y, x) of peaks
        refineds : numpy.ndarray
            numpy.ndarray of shape (n, 2) with float centers (y, x) of peaks (subpixel refinement)
        peak_values : numpy.ndarray
            numpy.ndarray of shape (n,) with float maxima of correlation map of peaks
        peak_elevations : numpy.ndarray
            numpy.ndarray of shape (n,) with float elevation of correlation map of peaks.
            See :meth:`libertem_blobfinder.base.correlation.peak_elevation` for details.
        zero : numpy.ndarray
            The near approximate zero point as numpy array (y, x).
        a,b : numpy.ndarray
            The near approximate vectors a, b to match the grid as numpy arrays (y, x).

        Returns
        -------

        Match
            :class:`~libertem.analysis.gridmatching.Match` object with the optimized
            matching result.
        '''
        corr = CorrelationResult(centers, refineds, peak_values, peak_elevations)
        filt = corr.peak_elevations >= self.min_weight

        selection = PointSelection(correlation_result=corr, selector=filt)
        # We match twice because we might catch more peaks in the second run with better parameters
        try:
            match1 = self._match_all(
                point_selection=selection, zero=zero, a=a, b=b
            )
            if len(match1) >= self.min_match:
                match1 = match1.weighted_optimize()
            else:
                raise np.linalg.LinAlgError("Not enough matched points")
            match2 = self._match_all(
                point_selection=selection, zero=match1.zero, a=match1.a, b=match1.b)
            return match2.weighted_optimize()
        except np.linalg.LinAlgError:
            return Match.invalid(corr)

    def affinematch(self, centers, indices, refineds=None, peak_values=None, peak_elevations=None):
        '''
        This function creates a Match object from correlation_result and
        indices for all points. The indices can be non-integer and relative to any
        base vectors zero, a, b, including virtual ones like zero=(0, 0), a=(1, 0), b=(0, 1).

        Refined values for zero, a and b that match the correlated peaks are then derived.

        This match method is very fast, can be robust against a distorted field of view and
        works without determining a lattice. It matches the full CorrelationResult and does
        not reject random points or other outliers.

        It is mathematically equivalent to calculating
        an affine transformation, as inspired by Giulio Guzzinati
        https://arxiv.org/abs/1902.06979

        Parameters
        ----------
        centers : numpy.ndarray
            numpy.ndarray of shape (n, 2) with integer centers (y, x) of peaks
        refineds : numpy.ndarray
            numpy.ndarray of shape (n, 2) with float centers (y, x) of peaks (subpixel refinement)
        peak_values : numpy.ndarray
            numpy.ndarray of shape (n,) with float maxima of correlation map of peaks
        peak_values : numpy.ndarray
            numpy.ndarray of shape (n,) with float elevation of correlation map of peaks.
            See :meth:`libertem_blobfinder.base.correlation.peak_elevation` for details.
        indices : numpy.ndarray
            The indices assigned to each point of the CorrelationResult.

        Returns
        -------
        Match
            :class:`~libertem.analysis.gridmatching.Match`
        '''
        corr = CorrelationResult(centers, refineds, peak_values, peak_elevations)
        match = Match(corr, selector=None, zero=None, a=None, b=None, indices=indices)
        try:
            return match.weighted_optimize()
        except np.linalg.LinAlgError:
            return Match.invalid(corr)

    def _match_all(self, point_selection: PointSelection, zero, a, b):
        '''
        Find points that can be generated from the lattice vectors with near integer indices

        Returns
        -------

        :class:`~libertem.analysis.gridmatching.Match`

        '''
        indices = get_indices(point_selection.refineds, zero, a, b)
        rounded = np.around(indices)
        index_diffs = np.absolute(indices - rounded)
        # We scale the difference from index dimension to the pixel dimension
        diffs = index_diffs * (np.linalg.norm(a), np.linalg.norm(b))
        # We scale far-out differences with the square root of the indices
        # to be more tolerant to small errors of a and b that result in large deviations
        # in absolute position at high indices
        scaled_diffs = diffs / (np.maximum(1, np.abs(indices))**0.5)
        errors = np.linalg.norm(scaled_diffs, axis=1)
        matched_selector = errors < self.tolerance
        matched_indices = rounded[matched_selector].astype(int)
        # remove the ones that weren't matched
        new_selector = point_selection.new_selector(matched_selector)
        result = Match.from_point_selection(
            point_selection, selector=new_selector, zero=zero, a=a, b=b, indices=matched_indices
        )
        return result


class Match(PointSelection):
    '''
    Class that represents a lattice match to a subset of a correlation result

    The attributes are not guaranteed to be correct or sensible for the given lattice.
    The methods :meth:`weighted_optimize` and :meth:`optimize`
    calculate a derived :class:`Match` with a best fit of :attr:`zero`,
    :attr:`a` and :attr:`b` based on the points and the indices.

    Attributes
    ----------

    zero : numpy.ndarray
        Declared zero point (y, x) of the lattice
    a : numpy.ndarray
        Declared "a" vector (y, x) of the lattice
    b : numpy.ndarray
        Declared "b" vector (y, x) of the lattice
    indices : numpy.ndarray
        List of indices (i, j) that are declared to express the matched points as linear combination
        of vectors :code:`a` and :code:`b` with reference to :code:`zero`. The indices
        can be integers or floats, and they can be precise or approximate, depending on the
        matching method.
    '''
    def __init__(self, correlation_result: CorrelationResult,
            selector, zero, a, b, indices):
        self.zero = zero
        self.a = a
        self.b = b
        self.indices = indices
        super().__init__(correlation_result, selector)
        assert len(indices) == len(self)

    def __str__(self):
        result = "zero: %s\n"\
            "a: %s\n"\
            "b: %s"
        return result % (str(self.zero), str(self.a), str(self.b))

    @classmethod
    def from_point_selection(cls, point_selection: PointSelection,
            zero, a, b, indices, selector=None):
        if selector is None:
            selector = point_selection.selector
        return Match(
            correlation_result=point_selection.correlation_result,
            selector=selector, zero=zero, a=a, b=b, indices=indices)

    @classmethod
    def invalid(cls, correlation_result):
        '''
        Match
            A :class:`Match` instance with empty selector and all-'nan' attributes
        '''
        nanvec = np.array([np.nan, np.nan])
        return cls(
            correlation_result=correlation_result,
            selector=np.zeros(len(correlation_result), dtype=bool),
            zero=nanvec,
            a=nanvec,
            b=nanvec,
            indices=np.array([]),
        )

    def isnan(self):
        return np.any(np.isnan(np.array([self.zero, self.a, self.b])))

    @property
    def calculated_refineds(self):
        '''
        numpy.ndarray : Calculated peak positions based on lattice parameters and indices.
        '''
        return calc_coords(self.zero, self.a, self.b, self.indices)

    def calc_coords(self, indices=None, drop_zero=False, frame_shape=None, r=0):
        '''
        Shorthand to calculate peak coordinates.

        Parameters
        ----------

        indices : numpy.ndarray
            Indices to calculate coordinates for. Both an array of (y, x) pairs
            and the output of np.mgrid are supported.
        drop_zero : bool
            Drop the zero order peak. This is important for virtual darkfield imaging.
        frame_shape : Tuple[int, int]
            If set, the peaks are filtered with :meth:`~libertem.analysis.gridmatching.within_frame`
        r : float
            Radius for :meth:`~libertem.analysis.gridmatching.within_frame`

        Returns
        -------

        numpy.ndarray
            A list of (y, x) coordinate pairs for peaks

        Raises
        ------

        ValueError
            If the shape of :code:`indices` is not as expected.
        '''
        if indices is None:
            indices = self.indices
        s = indices.shape
        # Output of mgrid
        if (len(s) == 3) and (s[0] == 2):
            indices = np.concatenate(indices.T)
        # List of (i, j) pairs
        elif (len(s) == 2) and (s[1] == 2):
            pass
        else:
            raise ValueError(
                "Shape of indices is %s, expected (n, 2) or (2, n, m)" % str(indices.shape))

        selector = np.ones(len(indices), dtype=bool)
        if drop_zero:
            nz = np.any(indices != 0, axis=1)
            selector *= nz
        peaks = calc_coords(self.zero, self.a, self.b, indices)
        if frame_shape is not None:
            fy, fx = frame_shape
            selector *= within_frame(peaks, r, fy, fx)
        return peaks[selector]

    @property
    def error(self):
        '''
        float : Weighted average distance between calculated and given peak position.
                numpy.float('inf') if match of length zero.
        '''
        if len(self) > 0:
            diff = np.linalg.norm(self.refineds - self.calculated_refineds, axis=1)
            return (diff * self.peak_elevations).mean() / self.peak_elevations.mean()
        else:
            return np.inf

    def derive(self, selector=None, zero=None, a=None, b=None, indices=None):
        if zero is None:
            zero = self.zero
        if a is None:
            a = self.a
        if b is None:
            b = self.b
        if indices is None:
            indices = self.indices
        if selector is None:
            selector = self.selector
        return Match(correlation_result=self.correlation_result, selector=selector,
            zero=zero, a=a, b=b, indices=indices)

    def weighted_optimize(self):
        '''
        Weighted least square optimization of :attr:`zero`, :attr:`a` and :attr:`b`

        Optimization to match the given points and indices using :attr:`peak_elevation` as weight.

        Returns
        -------

        Match
            A new :class:`Match` instance with optimized :attr:`zero`, :attr:`a` and :attr:`b`

        Raises
        ------

        np.linalg.LinAlgError
            If the solver didn't find a solution.
        '''

        # Following
        # https://stackoverflow.com/questions/27128688/how-to-use-least-squares-with-weight-matrix-in-python

        # We stack an index of 1 to the index list for the zero component
        indices = np.hstack([
            np.ones((len(self.indices), 1)),
            self.indices
        ])

        W = np.vstack([self.peak_elevations, self.peak_elevations])

        Aw = indices * np.sqrt(self.peak_elevations[:, np.newaxis])
        Bw = self.refineds * np.sqrt(W.T)
        (x, residuals, rank, s) = np.linalg.lstsq(
            Aw, Bw, rcond=None
        )
        # (zero, a, b)
        if x.size == 0:
            raise np.linalg.LinAlgError("Optimizing returned empty result")
        zero, a, b = x
        return self.derive(zero=zero, a=a, b=b)

    def optimize(self):
        '''
        Least square optimization of :attr:`zero`, :attr:`a` and :attr:`b`

        Optimization to match the given points and indices.

        Returns
        -------

        Match
            A new :class:`Match` instance with optimized :attr:`zero`, :attr:`a` and :attr:`b`

        Raises
        ------

        np.linalg.LinAlgError
            If the solver didn't find a solution.
        '''
        # We stack an index of 1 to the index list for the zero component
        indices = np.hstack([
            np.ones((len(self.indices), 1)),
            self.indices
        ])

        (x, residuals, rank, s) = np.linalg.lstsq(
            indices, self.refineds, rcond=None)
        # (zero, a, b)
        if x.size == 0:
            raise np.linalg.LinAlgError("Optimizing returned empty result")
        zero, a, b = x
        return self.derive(zero=zero, a=a, b=b)


def get_indices(points, zero, a, b):
    '''
    Find indices to express each point as sum of lattice vectors from zero point

    This could solve for arbitrarily many points, i.e. frame stacks instead of frame by frame
    With that the algorithm could actually match entire frame collections at once.
    '''
    coefficients = np.array((a, b)).T
    target = points - zero
    result = np.linalg.solve(coefficients, target.T).T
    return result


def get_transformation(ref, peaks, center=None, weighs=None):
    '''
    Inspired by Giulio Guzzinati
    https://arxiv.org/abs/1902.06979
    '''
    if center is None:
        center = np.array((0., 0.))

    assert ref.shape == peaks.shape
    A = np.hstack((ref - center, np.ones((len(ref), 1))))
    B = np.hstack((peaks - center, np.ones((len(peaks), 1))))

    if weighs is None:
        pass
    else:
        assert len(ref) == len(weighs)
        W = np.vstack((weighs, weighs, weighs)).T
        A *= W
        B *= W

    (fit, res, rank, s) = np.linalg.lstsq(A, B, rcond=None)
    return fit


def do_transformation(matrix, peaks, center=None):
    if center is None:
        center = np.array((0, 0))
    A = np.hstack((peaks - center, np.ones((len(peaks), 1))))
    B = np.dot(A, matrix)
    return B[:, 0:2] + center


def find_center(matrix):
    target = np.array((0, 0, 1)).T
    diff = np.identity(3)
    diff[2, 2] = 0
    # Find neutral point: solve a*m = a
    result = np.linalg.solve((matrix - diff).T, target)
    return result[0:2]



import pytest

from libertem.udf.base import UDFRunner, UDF


class UDF1Numpy(UDF):
    def get_result_buffers(self):
        return {}

    def process_frame(self):
        pass

    def get_backends(self):
        return ('numpy',)


class UDF2CuPy(UDF):
    def get_result_buffers(self):
        return {}

    def process_frame(self):
        pass

    def get_backends(self):
        return ('cupy',)


class UDF3NumpyCuPy(UDF):
    def get_result_buffers(self):
        return {}

    def process_frame(self):
        pass

    def get_backends(self):
        return ('cupy', 'numpy')


class UDF4CUDA(UDF):
    def get_result_buffers(self):
        return {}

    def process_frame(self):
        pass

    def get_backends(self):
        return ('cuda',)


class UDF5CUDAString(UDF):
    def get_result_buffers(self):
        return {}

    def process_frame(self):
        pass

    def get_backends(self):
        # Single string instead of tuple
        return 'cuda'


@pytest.mark.parametrize(
    "udfs",
    [(UDF1Numpy(), UDF2CuPy()), (UDF1Numpy(), UDF4CUDA())]
)
def test_no_common_backends(default_raw, udfs):
    runner = UDFRunner(udfs)
    tasks = list(runner._make_udf_tasks(
        dataset=default_raw, roi=None, backends=None,
    ))
    for task in tasks:
        with pytest.raises(ValueError) as e:
            task.get_resources()
        assert e.match("^There is no common supported UDF backend")


@pytest.mark.parametrize(
    "udfs,resources",
    [
        (
            [UDF1Numpy()],
            {'CPU': 1, 'compute': 1, 'ndarray': 1}
        ),
        (
            [UDF2CuPy()],
            {'CUDA': 1, 'compute': 1, 'ndarray': 1}
        ),
        (
            [UDF3NumpyCuPy()],
            {'compute': 1, 'ndarray': 1}
        ),
        (
            [UDF4CUDA()],
            {'CUDA': 1, 'compute': 1}
        ),
        (
            [UDF5CUDAString()],
            {'CUDA': 1, 'compute': 1}
        ),
        (
            [UDF1Numpy(), UDF3NumpyCuPy()],
            {'CPU': 1, 'compute': 1, 'ndarray': 1},
        ),
        (
            [UDF2CuPy(), UDF3NumpyCuPy()],
            {'CUDA': 1, 'compute': 1, 'ndarray': 1},
        ),
        (
            [UDF2CuPy(), UDF4CUDA()],
            {'CUDA': 1, 'compute': 1, 'ndarray': 1},
        ),
        (
            [UDF3NumpyCuPy(), UDF4CUDA()],
            {'CUDA': 1, 'compute': 1, 'ndarray': 1},
        ),
        (
            [UDF3NumpyCuPy(), UDF3NumpyCuPy()],
            {'compute': 1, 'ndarray': 1},
        ),
        (
            [UDF4CUDA(), UDF5CUDAString()],
            {'compute': 1, 'CUDA': 1},
        ),
    ]
)
def test_get_resources_combinations(default_raw, udfs, resources):
    runner = UDFRunner(udfs)
    tasks = list(runner._make_udf_tasks(
        dataset=default_raw, roi=None, backends=None,
    ))
    for task in tasks:
        assert task.get_resources() == resources
